class TestStatus:
    def test_status_leader(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        agent_self = c.agent.self()
        leader = c.status.leader()
        addr_port = agent_self["Stats"]["consul"]["leader_addr"]

        assert leader == addr_port, f"Leader value was {leader}, expected value was {addr_port}"

    def test_status_peers(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        agent_self = c.agent.self()

        addr_port = agent_self["Stats"]["consul"]["leader_addr"]
        peers = c.status.peers()

        assert addr_port in peers, f"Expected value '{addr_port}' in peer list but it was not present"



from __future__ import annotations

import json
from typing import Any, Optional

from consul import Check
from consul.callback import CB


class Agent:
    """
    The Agent endpoints are used to interact with a local Consul agent.
    Usually, services and checks are registered with an agent, which then
    takes on the burden of registering with the Catalog and performing
    anti-entropy to recover from outages.
    """

    def __init__(self, agent) -> None:
        self.agent = agent
        self.service = Agent.Service(agent)
        self.check = Agent.Check(agent)
        self.connect = Agent.Connect(agent)

    def self(self):
        """
        Returns configuration of the local agent and member information.
        """
        return self.agent.http.get(CB.json(), "/v1/agent/self")

    def services(self) -> Any:
        """
        Returns all the services that are registered with the local agent.
        These services were either provided through configuration files, or
        added dynamically using the HTTP API. It is important to note that
        the services known by the agent may be different than those
        reported by the Catalog. This is usually due to changes being made
        while there is no leader elected. The agent performs active
        anti-entropy, so in most situations everything will be in sync
        within a few seconds.
        """
        return self.agent.http.get(CB.json(), "/v1/agent/services")

    def service_definition(self, service_id):
        """
        Returns a service definition for a single instance that is registered
        with the local agent.
        """
        return self.agent.http.get(CB.json(), f"/v1/agent/service/{service_id}")

    def checks(self) -> Any:
        """
        Returns all the checks that are registered with the local agent.
        These checks were either provided through configuration files, or
        added dynamically using the HTTP API. Similar to services,
        the checks known by the agent may be different than those
        reported by the Catalog. This is usually due to changes being made
        while there is no leader elected. The agent performs active
        anti-entropy, so in most situations everything will be in sync
        within a few seconds.
        """
        return self.agent.http.get(CB.json(), "/v1/agent/checks")

    def members(self, wan: bool = False):
        """
        Returns all the members that this agent currently sees. This may
        vary by agent, use the nodes api of Catalog to retrieve a cluster
        wide consistent view of members.

        For agents running in server mode, setting *wan* to *True* returns
        the list of WAN members instead of the LAN members which is
        default.
        """
        params = []
        if wan:
            params.append(("wan", 1))
        return self.agent.http.get(CB.json(), "/v1/agent/members", params=params)

    def maintenance(self, enable: bool, reason: Optional[str] = None, token: str | None = None):
        """
        The node maintenance endpoint can place the agent into
        "maintenance mode".

        *enable* is either 'true' or 'false'. 'true' enables maintenance
        mode, 'false' disables maintenance mode.

        *reason* is an optional string. This is simply to aid human
        operators.
        """

        params: list[tuple[str, Any]] = []

        params.append(("enable", enable))
        if reason:
            params.append(("reason", reason))

        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(CB.boolean(), "/v1/agent/maintenance", params=params, headers=headers)

    def join(self, address: str, wan: bool = False, token: str | None = None):
        """
        This endpoint instructs the agent to attempt to connect to a
        given address.

        *address* is the ip to connect to.

        *wan* is either 'true' or 'false'. For agents running in server
        mode, 'true' causes the agent to attempt to join using the WAN
        pool. Default is 'false'.
        """

        params = []

        if wan:
            params.append(("wan", 1))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(CB.boolean(), f"/v1/agent/join/{address}", params=params, headers=headers)

    def force_leave(self, node: str, token: str | None = None):
        """
        This endpoint instructs the agent to force a node into the left
        state. If a node fails unexpectedly, then it will be in a failed
        state. Once in the failed state, Consul will attempt to reconnect,
        and the services and checks belonging to that node will not be
        cleaned up. Forcing a node into the left state allows its old
        entries to be removed.

        *node* is the node to change state for.
        """

        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(CB.boolean(), f"/v1/agent/force-leave/{node}", headers=headers)

    class Service:
        def __init__(self, agent) -> None:
            self.agent = agent

        def register(
            self,
            name: str,
            service_id=None,
            address=None,
            port: Optional[int] = None,
            tags=None,
            check=None,
            token: str | None = None,
            meta=None,
            weights=None,
            # *deprecated* use check parameter
            script=None,
            interval=None,
            ttl: Optional[int] = None,
            http=None,
            timeout=None,
            enable_tag_override: bool = False,
            extra_checks=None,
            replace_existing_checks=False,
        ):
            """
            Add a new service to the local agent. There is more
            documentation on services
            `here <http://www.consul.io/docs/agent/services.html>`_.

            *name* is the name of the service.

            If the optional *service_id* is not provided it is set to
            *name*. You cannot have duplicate *service_id* entries per
            agent, so it may be necessary to provide one.

            *address* will default to the address of the agent if not
            provided.

            An optional health *check* can be created for this service is
            one of `Check.script`_, `Check.http`_, `Check.tcp`_,
            `Check.ttl`_ or `Check.docker`_.

            *token* is an optional `ACL token`_ to apply to this request.
            Note this call will return successful even if the token doesn't
            have permissions to register this service.

            *meta* specifies arbitrary KV metadata linked to the service
            formatted as {k1:v1, k2:v2}.

            *weights* specifies weights for the service; default to
            {"Passing": 1, "Warning": 1}.

            *script*, *interval*, *ttl*, *http*, and *timeout* arguments
            are deprecated. use *check* instead.

            *replace_existing_checks* Missing health checks from the request will
            be deleted from the agent.
            Using this parameter allows to idempotently register a service and its
            checks without having to manually deregister checks.

            *enable_tag_override* is an optional bool that enable you
            to modify a service tags from servers(consul agent role server)
            Default is set to False.
            This option is only for >=v0.6.0 version on both agent and
            servers.
            for more information
            https://www.consul.io/docs/agent/services.html
            """

            if extra_checks is None:
                extra_checks = []
            payload: dict[str, Any] = {}

            payload["name"] = name
            if enable_tag_override:
                payload["enabletagoverride"] = enable_tag_override
            if service_id:
                payload["id"] = service_id
            if address:
                payload["address"] = address
            if port:
                payload["port"] = port
            if tags:
                payload["tags"] = tags
            if meta:
                payload["meta"] = meta
            if check:
                payload["checks"] = [check] + extra_checks
            if weights:
                payload["weights"] = weights
            else:
                payload.update(
                    Check._compat(  # pylint: disable=protected-access
                        script=script, interval=interval, ttl=ttl, http=http, timeout=timeout
                    )
                )
            params = []
            if replace_existing_checks:
                params.append(("replace-existing-checks", "true"))
            headers = self.agent.prepare_headers(token)
            return self.agent.http.put(
                CB.boolean(), "/v1/agent/service/register", params=params, headers=headers, data=json.dumps(payload)
            )

        def deregister(self, service_id: str, token: str | None = None):
            """
            Used to remove a service from the local agent. The agent will
            take care of deregistering the service with the Catalog. If
            there is an associated check, that is also deregistered.
            """
            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(CB.boolean(), f"/v1/agent/service/deregister/{service_id}", headers=headers)

        def maintenance(self, service_id: str, enable: bool, reason: Optional[str] = None, token: str | None = None):
            """
            The service maintenance endpoint allows placing a given service
            into "maintenance mode".

            *service_id* is the id of the service that is to be targeted
            for maintenance.

            *enable* is either 'true' or 'false'. 'true' enables
            maintenance mode, 'false' disables maintenance mode.

            *reason* is an optional string. This is simply to aid human
            operators.
            """

            params: list[tuple[str, Any]] = []

            params.append(("enable", enable))
            if reason:
                params.append(("reason", reason))

            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(
                CB.boolean(), f"/v1/agent/service/maintenance/{service_id}", params=params, headers=headers
            )

    class Check:
        def __init__(self, agent) -> None:
            self.agent = agent

        def register(
            self,
            name: str,
            check=None,
            check_id=None,
            notes=None,
            service_id=None,
            token: str | None = None,
            # *deprecated* use check parameter
            script=None,
            interval=None,
            ttl: Optional[int] = None,
            http=None,
            timeout=None,
        ):
            """
            Register a new check with the local agent. More documentation
            on checks can be found `here
            <http://www.consul.io/docs/agent/checks.html>`_.

            *name* is the name of the check.

            *check* is one of `Check.script`_, `Check.http`_, `Check.tcp`_
            `Check.ttl`_ or `Check.docker`_ and is required.

            If the optional *check_id* is not provided it is set to *name*.
            *check_id* must be unique for this agent.

            *notes* is not used by Consul, and is meant to be human
            readable.

            Optionally, a *service_id* can be specified to associate a
            registered check with an existing service.

            *token* is an optional `ACL token`_ to apply to this request.
            Note this call will return successful even if the token doesn't
            have permissions to register this check.

            *script*, *interval*, *ttl*, *http*, and *timeout* arguments
            are deprecated. use *check* instead.

            Returns *True* on success.
            """
            payload = {"name": name}

            assert check or script or ttl or http, "check is required"

            if check:
                payload.update(check)

            else:
                payload.update(
                    Check._compat(script=script, interval=interval, ttl=ttl, http=http, timeout=timeout)["check"]
                )

            if check_id:
                payload["id"] = check_id
            if notes:
                payload["notes"] = notes
            if service_id:
                payload["serviceid"] = service_id

            headers = self.agent.prepare_headers(token)
            return self.agent.http.put(
                CB.boolean(), "/v1/agent/check/register", headers=headers, data=json.dumps(payload)
            )

        def deregister(self, check_id: str, token: str | None = None):
            """
            Remove a check from the local agent.
            """
            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(CB.boolean(), f"/v1/agent/check/deregister/{check_id}", headers=headers)

        def ttl_pass(self, check_id: str, notes=None, token: str | None = None):
            """
            Mark a ttl based check as passing. Optional notes can be
            attached to describe the status of the check.
            """
            params = []
            if notes:
                params.append(("note", notes))
            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(CB.boolean(), f"/v1/agent/check/pass/{check_id}", params=params, headers=headers)

        def ttl_fail(self, check_id: str, notes=None, token: str | None = None):
            """
            Mark a ttl based check as failing. Optional notes can be
            attached to describe why check is failing. The status of the
            check will be set to critical and the ttl clock will be reset.
            """
            params = []
            if notes:
                params.append(("note", notes))
            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(CB.boolean(), f"/v1/agent/check/fail/{check_id}", params=params, headers=headers)

        def ttl_warn(self, check_id: str, notes=None, token: str | None = None):
            """
            Mark a ttl based check with warning. Optional notes can be
            attached to describe the warning. The status of the
            check will be set to warn and the ttl clock will be reset.
            """
            params = []
            if notes:
                params.append(("note", notes))
            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(CB.boolean(), f"/v1/agent/check/warn/{check_id}", params=params, headers=headers)

    class Connect:
        def __init__(self, agent) -> None:
            self.agent = agent
            self.ca = Agent.Connect.CA(agent)

        def authorize(self, target, client_cert_uri, client_cert_serial, token: str | None = None):
            """
            Tests whether a connection attempt is authorized between
            two services.
            More information is available
            `here <https://www.consul.io/api-docs/agent/connect>`_.

            *target* is the name of the service that is being requested.

            *client_cert_uri* The unique identifier for the requesting
            client.

            *client_cert_serial* The colon-hex-encoded serial number for
            the requesting client cert.
            """

            payload = {"Target": target, "ClientCertURI": client_cert_uri, "ClientCertSerial": client_cert_serial}

            headers = self.agent.prepare_headers(token)

            return self.agent.http.put(
                CB.json(), "/v1/agent/connect/authorize", headers=headers, data=json.dumps(payload)
            )

        class CA:
            def __init__(self, agent) -> None:
                self.agent = agent

            def roots(self):
                return self.agent.http.get(CB.json(), "/v1/agent/connect/ca/roots")

            def leaf(self, service, token: str | None = None):
                headers = self.agent.prepare_headers(token)

                return self.agent.http.get(CB.json(), f"/v1/agent/connect/ca/leaf/{service}", headers=headers)



import json

from consul.callback import CB


class Txn:
    """
    The Transactions endpoints manage updates or fetches of multiple keys
    inside a single, atomic transaction.
    """

    def __init__(self, agent) -> None:
        self.agent = agent

    def put(self, payload):
        """
        Create a transaction by submitting a list of operations to apply to
        the KV store inside of a transaction. If any operation fails, the
        transaction is rolled back and none of the changes are applied.

        *payload* is a list of operations where each operation is a `dict`
        with a single key value pair, with the key specifying operation the
        type. An example payload of operation type "KV" is
        dict::

            {
                "KV": {
                  "Verb": "<verb>",
                  "Key": "<key>",
                  "Value": "<Base64-encoded blob of data>",
                  "Flags": 0,
                  "Index": 0,
                  "Session": "<session id>"
                }
            }
        """
        return self.agent.http.put(CB.json(), "/v1/txn", data=json.dumps(payload))



from consul.callback import CB


class Status:
    """
    The Status endpoints are used to get information about the status
     of the Consul cluster.
    """

    def __init__(self, agent) -> None:
        self.agent = agent

    def leader(self):
        """
        This endpoint is used to get the Raft leader for the datacenter
        in which the agent is running.
        """
        return self.agent.http.get(CB.json(), "/v1/status/leader")

    def peers(self):
        """
        This endpoint retrieves the Raft peers for the datacenter in which
        the the agent is running.
        """
        return self.agent.http.get(CB.json(), "/v1/status/peers")



from __future__ import annotations

import json
import typing

from consul.callback import CB

if typing.TYPE_CHECKING:
    import builtins


class Token:
    def __init__(self, agent) -> None:
        self.agent = agent

    def list(self, token: str | None = None):
        """
        Lists all the active ACL tokens. This is a privileged endpoint, and
        requires a management token. *token* will override this client's
        default token.
        Requires a token with acl:read capability. ACLPermissionDenied raised otherwise
        """
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), "/v1/acl/tokens", headers=headers)

    def read(self, accessor_id: str, token: str | None = None):
        """
        Returns the token information for *accessor_id*. Requires a token with acl:read capability.
        :param accessor_id: The accessor ID of the token to read
        :param token: token with acl:read capability
        :return: selected token information
        """
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), f"/v1/acl/token/{accessor_id}", headers=headers)

    def delete(self, accessor_id: str, token: str | None = None):
        """
        Deletes the token with *accessor_id*. This is a privileged endpoint, and requires a token with acl:write.
        :param accessor_id: The accessor ID of the token to delete
        :param token: token with acl:write capability
        :return: True if the token was deleted
        """
        headers = self.agent.prepare_headers(token)
        return self.agent.http.delete(CB.boolean(), f"/v1/acl/token/{accessor_id}", headers=headers)

    def clone(self, accessor_id: str, token: str | None = None, description: str = ""):
        """
        Clones the token identified by *accessor_id*. This is a privileged endpoint, and requires a token with acl:write.
        :param accessor_id: The accessor ID of the token to clone
        :param token: token with acl:write capability
        :param description: Optional new token description
        :return: The cloned token information
        """

        json_data = {"Description": description}
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.json(),
            f"/v1/acl/token/{accessor_id}/clone",
            headers=headers,
            data=json.dumps(json_data),
        )

    def create(
        self,
        token: str | None = None,
        accessor_id: str | None = None,
        secret_id: str | None = None,
        policies_id: builtins.list[str] | None = None,
        description: str = "",
    ):
        """
        Create a token (optionally identified by *secret_id* and *accessor_id*).
        This is a privileged endpoint, and requires a token with acl:write.
        :param token: token with acl:write capability
        :param accessor_id: The accessor ID of the token to create
        :param secret_id: The secret ID of the token to create
        :param description: Optional new token description
        :param policies_id: Optional list of policies id
        :return: The cloned token information
        """

        json_data: dict[str, typing.Any] = {}
        if accessor_id:
            json_data["AccessorID"] = accessor_id
        if secret_id:
            json_data["SecretID"] = secret_id
        if description:
            json_data["Description"] = description
        if policies_id:
            json_data["Policies"] = [{"ID": policy} for policy in policies_id]

        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.json(),
            "/v1/acl/token",
            headers=headers,
            data=json.dumps(json_data),
        )

    def update(self, accessor_id: str, token: str | None = None, secret_id: str | None = None, description: str = ""):
        """
        Update a token (optionally identified by *secret_id* and *accessor_id*).
        This is a privileged endpoint, and requires a token with acl:write.
        :param accessor_id: The accessor ID of the token to update
        :param token: token with acl:write capability
        :param secret_id: Optional secret ID of the token to update
        :param description: Optional new token description
        :return: The updated token information
        """

        json_data = {"AccessorID": accessor_id}
        if secret_id:
            json_data["SecretID"] = secret_id
        if description:
            json_data["Description"] = description
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.json(),
            f"/v1/acl/token/{accessor_id}",
            headers=headers,
            data=json.dumps(json_data),
        )



from __future__ import annotations

from packaging import version


def find_recursive(list_or_single_dict: list[dict] | dict, wanted: list[dict] | dict) -> bool:
    """
    Check if the 'wanted' item(s) sub-dictionary structure is/are present in any of the dictionaries in 'list_or_single_dict'.
    'list_or_single_dict' can be either a list of dictionaries or a single dictionary.
    """

    def matches_subdict(haystack: dict, needle: dict) -> bool:
        """
        Check if all key-value pairs in needle are present in haystack.
        This function recursively checks nested dictionaries and lists of dictionaries.
        """
        for key, needle_value in needle.items():
            if key not in haystack:
                return False

            haystack_value = haystack[key]

            if isinstance(needle_value, dict):
                if not isinstance(haystack_value, dict) or not matches_subdict(haystack_value, needle_value):
                    return False
            elif isinstance(needle_value, list):
                if not all(
                    any(matches_subdict(haystack_item, needle_item) for haystack_item in haystack_value)
                    for needle_item in needle_value
                ):
                    return False
            else:
                if haystack_value != needle_value:
                    return False

        return True

    # Turn both args into lists to handle single/multi lookup
    if isinstance(list_or_single_dict, dict):
        list_or_single_dict = [list_or_single_dict]

    if isinstance(wanted, dict):
        wanted = [wanted]

    return all(any(matches_subdict(item, single_wanted) for item in list_or_single_dict) for single_wanted in wanted)


def should_skip(version_str: str, comparator: str, ref_version_str: str) -> bool:
    v = version.parse(version_str)
    ref_version = version.parse(ref_version_str)
    if (
        (comparator == "<=" and v > ref_version)  # pylint: disable=too-many-boolean-expressions
        or (comparator == ">=" and v < ref_version)
        or (comparator == "<" and v >= ref_version)
        or (comparator == ">" and v <= ref_version)
    ):
        print(f"Requires version {comparator} {ref_version_str}. Got {v}")
        return True

    return False



import time

import consul.check

Check = consul.check.Check


class TestHealth:
    def test_health_service(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        # check there are no nodes for the service 'foo'
        _index, nodes = c.health.service("foo")
        assert nodes == []

        # register two nodes, one with a long ttl, the other shorter
        c.agent.service.register("foo", service_id="foo:1", check=Check.ttl("10s"), tags=["tag:foo:1"])
        c.agent.service.register("foo", service_id="foo:2", check=Check.ttl("1s"))

        time.sleep(0.2)

        # check the nodes show for the /health/service endpoint
        _index, nodes = c.health.service("foo")
        assert [node["Service"]["ID"] for node in nodes] == ["foo:1", "foo:2"]

        # but that they aren't passing their health check
        _index, nodes = c.health.service("foo", passing=True)
        assert nodes == []

        # ping the two node's health check
        c.agent.check.ttl_pass("service:foo:1")
        c.agent.check.ttl_pass("service:foo:2")

        time.sleep(0.2)

        # both nodes are now available
        _index, nodes = c.health.service("foo", passing=True)
        assert [node["Service"]["ID"] for node in nodes] == ["foo:1", "foo:2"]

        # wait until the short ttl node fails
        time.sleep(3)

        # only one node available
        _index, nodes = c.health.service("foo", passing=True)
        assert [node["Service"]["ID"] for node in nodes] == ["foo:1"]

        # ping the failed node's health check
        c.agent.check.ttl_pass("service:foo:2")

        time.sleep(0.2)

        # check both nodes are available
        _index, nodes = c.health.service("foo", passing=True)
        assert [node["Service"]["ID"] for node in nodes] == ["foo:1", "foo:2"]

        # check that tag works
        _index, nodes = c.health.service("foo", tag="tag:foo:1")
        assert [node["Service"]["ID"] for node in nodes] == ["foo:1"]

        # deregister the nodes
        c.agent.service.deregister("foo:1")
        c.agent.service.deregister("foo:2")

        time.sleep(0.2)

        _index, nodes = c.health.service("foo")
        assert nodes == []

    def test_health_state(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        # The empty string is for the Serf Health Status check, which has an
        # empty ServiceID
        _index, nodes = c.health.state("any")
        assert [node["ServiceID"] for node in nodes] == [""]

        # register two nodes, one with a long ttl, the other shorter
        c.agent.service.register("foo", service_id="foo:1", check=Check.ttl("10s"))
        c.agent.service.register("foo", service_id="foo:2", check=Check.ttl("1s"))

        time.sleep(0.2)

        # check the nodes show for the /health/state/any endpoint
        _index, nodes = c.health.state("any")
        assert {node["ServiceID"] for node in nodes} == {"", "foo:1", "foo:2"}

        # but that they aren't passing their health check
        _index, nodes = c.health.state("passing")
        assert "foo" not in [node["ServiceID"] for node in nodes]

        # ping the two node's health check
        c.agent.check.ttl_pass("service:foo:1")
        c.agent.check.ttl_pass("service:foo:2")

        time.sleep(0.2)

        # both nodes are now available
        _index, nodes = c.health.state("passing")
        assert {node["ServiceID"] for node in nodes} == {"", "foo:1", "foo:2"}

        # wait until the short ttl node fails
        time.sleep(3)

        # only one node available
        _index, nodes = c.health.state("passing")
        assert {node["ServiceID"] for node in nodes} == {"", "foo:1"}

        # ping the failed node's health check
        c.agent.check.ttl_pass("service:foo:2")

        time.sleep(0.2)

        # check both nodes are available
        _index, nodes = c.health.state("passing")
        assert {node["ServiceID"] for node in nodes} == {"", "foo:1", "foo:2"}

        # deregister the nodes
        c.agent.service.deregister("foo:1")
        c.agent.service.deregister("foo:2")

        time.sleep(0.2)

        _index, nodes = c.health.state("any")
        assert [node["ServiceID"] for node in nodes] == [""]

    def test_health_node(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        # grab local node name
        node = c.agent.self()["Config"]["NodeName"]
        _index, checks = c.health.node(node)
        assert node in [check["Node"] for check in checks]

    def test_health_checks(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        c.agent.service.register("foobar", service_id="foobar", check=Check.ttl("10s"))

        time.sleep(40 / 1000.00)

        _index, checks = c.health.checks("foobar")

        assert [check["ServiceID"] for check in checks] == ["foobar"]
        assert [check["CheckID"] for check in checks] == ["service:foobar"]

        c.agent.service.deregister("foobar")

        time.sleep(40 / 1000.0)

        _index, checks = c.health.checks("foobar")
        assert len(checks) == 0



from __future__ import annotations

from typing import Optional

from consul.callback import CB


class Event:
    """
    The event command provides a mechanism to fire a custom user event to
    an entire datacenter. These events are opaque to Consul, but they can
    be used to build scripting infrastructure to do automated deploys,
    restart services, or perform any other orchestration action.

    Unlike most Consul data, which is replicated using consensus, event
    data is purely peer-to-peer over gossip.

    This means it is not persisted and does not have a total ordering. In
    practice, this means you cannot rely on the order of message delivery.
    An advantage however is that events can still be used even in the
    absence of server nodes or during an outage."""

    def __init__(self, agent) -> None:
        self.agent = agent

    def fire(self, name: str, body: str = "", node=None, service=None, tag=None, token: str | None = None):
        """
        Sends an event to Consul's gossip protocol.

        *name* is the Consul-opaque name of the event. This can be filtered
        on in calls to list, below

        *body* is the Consul-opaque body to be delivered with the event.
         From the Consul documentation:
            The underlying gossip also sets limits on the size of a user
            event message. It is hard to give an exact number, as it
            depends on various parameters of the event, but the payload
            should be kept very small (< 100 bytes²). Specifying too large
            of an event will return an error.

        *node*, *service*, and *tag* are regular expressions which remote
        agents will filter against to determine if they should store the
        event

        *token* is an optional `ACL token`_ to apply to this request. If
        the token's policy is not allowed to fire an event of this *name*
        an *ACLPermissionDenied* exception will be raised.
        """
        assert not name.startswith("/"), "keys should not start with a forward slash"
        params = []
        if node is not None:
            params.append(("node", node))
        if service is not None:
            params.append(("service", service))
        if tag is not None:
            params.append(("tag", tag))

        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(CB.json(), f"/v1/event/fire/{name}", params=params, headers=headers, data=body)

    def list(self, name: Optional[str] = None, index=None, wait=None):
        """
        Returns a tuple of (*index*, *events*)
            Note: Since Consul's event protocol uses gossip, there is no
            ordering, and instead index maps to the newest event that
            matches the query.

        *name* is the type of events to list, if None, lists all available.

        *index* is the current event Consul index, suitable for making
        subsequent calls to wait for changes since this query was last run.
        Check https://consul.io/docs/agent/http/event.html#event_list for
        more infos about indexes on events.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. This parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        Consul agents only buffer the most recent entries. The current
        buffer size is 256, but this value could change in the future.

        Each *event* looks like this::

            {
                  {
                    "ID": "b54fe110-7af5-cafc-d1fb-afc8ba432b1c",
                    "Name": "deploy",
                    "Payload": "1609030",
                    "NodeFilter": "",
                    "ServiceFilter": "",
                    "TagFilter": "",
                    "Version": 1,
                    "LTime": 19
                  },
            }
        """
        params = []
        if name is not None:
            params.append(("name", name))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        return self.agent.http.get(CB.json(index=True, decode="Payload"), "/v1/event/list", params=params)



from __future__ import annotations

from consul.callback import CB


class KV:
    """
    The KV endpoint is used to expose a simple key/value store. This can be
    used to store service configurations or other meta data in a simple
    way.
    """

    def __init__(self, agent) -> None:
        self.agent = agent

    def get(
        self,
        key,
        index=None,
        recurse: bool = False,
        wait=None,
        token: str | None = None,
        consistency=None,
        keys: bool = False,
        separator=None,
        dc=None,
        connections_timeout=None,
    ):
        """
        Returns a tuple of (*index*, *value[s]*)

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *token* is an optional `ACL token`_ to apply to this request.

        *keys* is a boolean which, if True, says to return a flat list of
        keys without values or other metadata. *separator* can be used
        with *keys* to list keys only up to a given separator character.

        *dc* is the optional datacenter that you wish to communicate with.
        If None is provided, defaults to the agent's datacenter.

        The *value* returned is for the specified key, or if *recurse* is
        True a list of *values* for all keys with the given prefix is
        returned.

        Each *value* looks like this::

            {
                "CreateIndex": 100,
                "ModifyIndex": 200,
                "LockIndex": 200,
                "Key": "foo",
                "Flags": 0,
                "Value": "bar",
                "Session": "adf4238a-882b-9ddc-4a9d-5b6758e4159e"
            }

        Note, if the requested key does not exists *(index, None)* is
        returned. It's then possible to long poll on the index for when the
        key is created.
        """
        assert not key.startswith("/"), "keys should not start with a forward slash"
        params = []
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        if recurse:
            params.append(("recurse", "1"))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if keys:
            params.append(("keys", True))
        if separator:
            params.append(("separator", separator))
        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))

        one = False
        decode: bool | str = False

        if not keys:
            decode = "Value"
        if not recurse and not keys:
            one = True
        http_kwargs = {}
        if connections_timeout:
            http_kwargs["connections_timeout"] = connections_timeout

        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(
            CB.json(index=True, decode=decode, one=one), f"/v1/kv/{key}", params=params, headers=headers, **http_kwargs
        )

    def put(
        self,
        key,
        value,
        cas=None,
        flags=None,
        acquire=None,
        release=None,
        token: str | None = None,
        dc=None,
        connections_timeout=None,
    ):
        """
        Sets *key* to the given *value*.

        *value* can either be None (useful for marking a key as a
        directory) or any string type, including binary data (e.g. a
        msgpack'd data structure)

        The optional *cas* parameter is used to turn the PUT into a
        Check-And-Set operation. This is very useful as it allows clients
        to build more complex syncronization primitives on top. If the
        index is 0, then Consul will only put the key if it does not
        already exist. If the index is non-zero, then the key is only set
        if the index matches the ModifyIndex of that key.

        An optional *flags* can be set. This can be used to specify an
        unsigned value between 0 and 2^64-1.

        *acquire* is an optional session_id. if supplied a lock acquisition
        will be attempted.

        *release* is an optional session_id. if supplied a lock release
        will be attempted.

        *token* is an optional `ACL token`_ to apply to this request. If
        the token's policy is not allowed to write to this key an
        *ACLPermissionDenied* exception will be raised.

        *dc* is the optional datacenter that you wish to communicate with.
        If None is provided, defaults to the agent's datacenter.

        The return value is simply either True or False. If False is
        returned, then the update has not taken place.
        """
        assert not key.startswith("/"), "keys should not start with a forward slash"
        assert value is None or isinstance(value, (str, bytes)), "value should be None or a string / binary data"

        params = []
        if cas is not None:
            params.append(("cas", cas))
        if flags is not None:
            params.append(("flags", flags))
        if acquire:
            params.append(("acquire", acquire))
        if release:
            params.append(("release", release))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        http_kwargs = {}
        if connections_timeout:
            http_kwargs["connections_timeout"] = connections_timeout
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.json(), f"/v1/kv/{key}", params=params, headers=headers, data=value, **http_kwargs
        )

    def delete(self, key, recurse=None, cas=None, token: str | None = None, dc=None, connections_timeout=None):
        """
        Deletes a single key or if *recurse* is True, all keys sharing a
        prefix.

        *cas* is an optional flag is used to turn the DELETE into a
        Check-And-Set operation. This is very useful as a building block
        for more complex synchronization primitives. Unlike PUT, the index
        must be greater than 0 for Consul to take any action: a 0 index
        will not delete the key. If the index is non-zero, the key is only
        deleted if the index matches the ModifyIndex of that key.

        *token* is an optional `ACL token`_ to apply to this request. If
        the token's policy is not allowed to delete to this key an
        *ACLPermissionDenied* exception will be raised.

        *dc* is the optional datacenter that you wish to communicate with.
        If None is provided, defaults to the agent's datacenter.
        """
        assert not key.startswith("/"), "keys should not start with a forward slash"

        params = []
        if recurse:
            params.append(("recurse", "1"))
        if cas is not None:
            params.append(("cas", cas))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        http_kwargs = {}
        if connections_timeout:
            http_kwargs["connections_timeout"] = connections_timeout
        headers = self.agent.prepare_headers(token)
        return self.agent.http.delete(CB.json(), f"/v1/kv/{key}", params=params, headers=headers, **http_kwargs)



import collections

import pytest

from consul import Consul

ACLConsul = collections.namedtuple("ACLConsul", ["instance", "token", "version"])


@pytest.fixture
def acl_consul(acl_consul_instance):
    instance, token = acl_consul_instance
    return ACLConsul(Consul(port=instance.port), token, instance.version)


@pytest.fixture
def consul_obj(consul_port):
    consul_port, consul_version = consul_port
    c = Consul(port=consul_port)
    return c, consul_version



from typing import Optional

import requests
from requests import Response

from consul import base

__all__ = ["Consul"]


class HTTPClient(base.HTTPClient):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.session = requests.session()

    def response(self, response: Response):
        response.encoding = "utf-8"
        return base.Response(response.status_code, response.headers, response.text)

    def get(self, callback, path, params=None, headers: Optional[dict[str, str]] = None):
        uri = self.uri(path, params)
        return callback(self.response(self.session.get(uri, headers=headers, verify=self.verify, cert=self.cert)))

    def put(self, callback, path, params=None, data: str = "", headers: Optional[dict[str, str]] = None):
        uri = self.uri(path, params)
        return callback(
            self.response(self.session.put(uri, headers=headers, data=data, verify=self.verify, cert=self.cert))
        )

    def delete(self, callback, path, params=None, headers: Optional[dict[str, str]] = None):
        uri = self.uri(path, params)
        return callback(self.response(self.session.delete(uri, headers=headers, verify=self.verify, cert=self.cert)))

    def post(self, callback, path, params=None, data: str = "", headers: Optional[dict[str, str]] = None):
        uri = self.uri(path, params)
        return callback(
            self.response(self.session.post(uri, headers=headers, data=data, verify=self.verify, cert=self.cert))
        )

    def close(self) -> None:
        pass


class Consul(base.Consul):
    def http_connect(self, host: str, port: int, scheme, verify: bool = True, cert=None):
        return HTTPClient(host, port, scheme, verify, cert)



from consul.callback import CB


class Coordinate:
    def __init__(self, agent) -> None:
        self.agent = agent

    def datacenters(self):
        """
        Returns the WAN network coordinates for all Consul servers,
        organized by DCs.
        """
        return self.agent.http.get(CB.json(), "/v1/coordinate/datacenters")

    def nodes(self, dc=None, index=None, wait=None, consistency=None):
        """
        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *consistency* can be either 'default', 'consistent' or 'stale'. if
        not specified *consistency* will the consistency level this client
        was configured with.
        """
        params = []
        if dc:
            params.append(("dc", dc))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))
        return self.agent.http.get(CB.json(index=True), "/v1/coordinate/nodes", params=params)



from consul.callback import CB


class Operator:
    def __init__(self, agent) -> None:
        self.agent = agent

    def raft_config(self):
        """
        Returns raft configuration.
        """
        return self.agent.http.get(CB.json(), "/v1/operator/raft/configuration")



from tests.utils import find_recursive, should_skip


def test_find_recursive() -> None:
    ret_value = [
        {
            "AccessorID": "accessorid",
            "CreateIndex": 1,
            "CreateTime": "timestamp",
            "Description": "description",
            "Hash": "hash",
            "Local": False,
            "ModifyIndex": 1,
            "Policies": [{"ID": "id", "Name": "name"}],
            "SecretID": "secretid",
        },
        {
            "AccessorID": "accessorid",
            "CreateIndex": 1,
            "CreateTime": "timestamp",
            "Description": "description",
            "Hash": "hash",
            "Local": False,
            "ModifyIndex": 1,
            "Policies": [{"ID": "id", "Name": "name"}],
            "SecretID": "secretid2",
        },
    ]

    wanted = {
        "AccessorID": "accessorid",
        "Description": "description",
        "Policies": [{"Name": "name"}],
        "SecretID": "secretid",
    }
    wanted2 = {
        "AccessorID": "accessorid",
        "SecretID": "secretid2",
    }
    unwanted = {
        "AccessorID": "accessorid",
        "Description": "description",
        "Policies": [{"Name": "name-ish"}],
        "SecretID": "secretid",
    }

    assert find_recursive(ret_value, wanted)
    assert find_recursive(ret_value, wanted2)
    assert find_recursive(ret_value, [wanted, wanted2])
    assert find_recursive(wanted, wanted)
    assert not find_recursive(ret_value, unwanted)


def test_should_skip() -> None:
    test_cases = [
        ("1.0.0", "<=", "1.0.0", False),
        ("1.0.1", "<=", "1.0.0", True),
        ("0.9.9", "<=", "1.0.0", False),
        ("1.0.0", ">=", "1.0.0", False),
        ("1.0.1", ">=", "1.0.0", False),
        ("0.9.9", ">=", "1.0.0", True),
        ("1.0.0", "<", "1.0.0", True),
        ("1.0.1", "<", "1.0.0", True),
        ("0.9.9", "<", "1.0.0", False),
        ("1.0.0", ">", "1.0.0", True),
        ("1.0.1", ">", "1.0.0", False),
        ("0.9.9", ">", "1.0.0", True),
    ]

    for version_str, comparator, ref_version_str, expected in test_cases:
        assert should_skip(version_str, comparator, ref_version_str) == expected



from __future__ import annotations

import logging
import warnings
from typing import Any, Optional

log = logging.getLogger(__name__)


class Check:
    """
    There are three different kinds of checks: script, http and ttl
    """

    @classmethod
    def script(cls, args, interval, deregister=None) -> dict[str, Any]:
        """
        Run the script *args* every *interval* (e.g. "10s") to peform health
        check
        """
        if isinstance(args, (str, bytes)):
            warnings.warn("Check.script should take a list of args", DeprecationWarning)
            args = ["sh", "-c", args]
        ret = {"args": args, "interval": interval}
        if deregister:
            ret["DeregisterCriticalServiceAfter"] = deregister
        return ret

    @classmethod
    def http(cls, url, interval, timeout=None, deregister=None, header=None) -> dict[str, Any]:
        """
        Peform a HTTP GET against *url* every *interval* (e.g. "10s") to peform
        health check with an optional *timeout* and optional *deregister* after
        which a failing service will be automatically deregistered. Optional
        parameter *header* specifies headers sent in HTTP request. *header*
        paramater is in form of map of lists of strings,
        e.g. {"x-foo": ["bar", "baz"]}.
        """
        ret = {"http": url, "interval": interval}
        if timeout:
            ret["timeout"] = timeout
        if deregister:
            ret["DeregisterCriticalServiceAfter"] = deregister
        if header:
            ret["header"] = header
        return ret

    @classmethod
    def tcp(cls, host: str, port: int, interval, timeout=None, deregister=None) -> dict[str, Any]:
        """
        Attempt to establish a tcp connection to the specified *host* and
        *port* at a specified *interval* with optional *timeout* and optional
        *deregister* after which a failing service will be automatically
        deregistered.
        """
        ret = {"tcp": f"{host:s}:{port:d}", "interval": interval}
        if timeout:
            ret["timeout"] = timeout
        if deregister:
            ret["DeregisterCriticalServiceAfter"] = deregister
        return ret

    @classmethod
    def ttl(cls, ttl: str) -> dict[str, Any]:
        """
        Set check to be marked as critical after *ttl* (e.g. "10s") unless the
        check is periodically marked as passing.
        """
        return {"ttl": ttl}

    @classmethod
    def docker(cls, container_id, shell, script, interval, deregister=None) -> dict[str, Any]:
        """
        Invoke *script* packaged within a running docker container with
        *container_id* at a specified *interval* on the configured
        *shell* using the Docker Exec API.  Optional *register* after which a
        failing service will be automatically deregistered.
        """
        ret = {"docker_container_id": container_id, "shell": shell, "script": script, "interval": interval}
        if deregister:
            ret["DeregisterCriticalServiceAfter"] = deregister
        return ret

    @classmethod
    def _compat(
        cls, script=None, interval=None, ttl: Optional[int] = None, http=None, timeout=None, deregister=None
    ) -> dict[str, Any]:
        if not script and not http and not ttl:
            return {}

        log.warning("DEPRECATED: use consul.Check.script/http/ttl to specify check")

        ret: dict[str, Any] = {"check": {}}

        if script:
            assert interval
            assert not ttl
            assert not http
            ret["check"] = {"script": script, "interval": interval}
        if ttl:
            assert not interval or script
            assert not http
            ret["check"] = {"ttl": ttl}
        if http:
            assert interval
            assert not script
            assert not ttl
            ret["check"] = {"http": http, "interval": interval}
        if timeout:
            assert http
            ret["check"]["timeout"] = timeout

        if deregister:
            ret["check"]["DeregisterCriticalServiceAfter"] = deregister

        return ret



from __future__ import annotations

import abc
import collections
import logging
import os
import urllib
from typing import TYPE_CHECKING, Any, Optional

from consul.api.acl import ACL
from consul.api.agent import Agent
from consul.api.catalog import Catalog
from consul.api.connect import Connect
from consul.api.coordinates import Coordinate
from consul.api.event import Event
from consul.api.health import Health
from consul.api.kv import KV
from consul.api.operator import Operator
from consul.api.query import Query
from consul.api.session import Session
from consul.api.status import Status
from consul.api.txn import Txn
from consul.exceptions import ConsulException

if TYPE_CHECKING:
    from types import TracebackType

log = logging.getLogger(__name__)


#
# Convenience to define checks


Response = collections.namedtuple("Response", ["code", "headers", "body"])


class HTTPClient(metaclass=abc.ABCMeta):
    def __init__(
        self, host: str = "127.0.0.1", port: int = 8500, scheme: str = "http", verify: bool = True, cert=None
    ) -> None:
        self.host = host
        self.port = port
        self.scheme = scheme
        self.verify = verify
        self.base_uri = f"{self.scheme}://{self.host}:{self.port}"
        self.cert = cert

    def uri(self, path: str, params: list[tuple[str, Any]] | None = None):
        uri = self.base_uri + urllib.parse.quote(path, safe="/:")
        if params:
            uri = f"{uri}?{urllib.parse.urlencode(params)}"
        return uri

    @abc.abstractmethod
    def get(self, callback, path, params=None, headers: Optional[dict[str, str]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, callback, path, params=None, data: str = "", headers: Optional[dict[str, str]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, callback, path, params=None, headers: Optional[dict[str, str]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def post(self, callback, path, params=None, data: str = "", headers: Optional[dict[str, str]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError


class Consul:
    def __init__(
        self,
        host: str | None = None,
        port: int | None = None,
        token: str | None = None,
        scheme: str | None = None,
        consistency: str = "default",
        dc=None,
        verify: bool | None = None,
        cert=None,
    ) -> None:
        """
        *token* is an optional `ACL token`_. If supplied it will be used by
        default for all requests made with this client session. It's still
        possible to override this token by passing a token explicitly for a
        request.

        *consistency* sets the consistency mode to use by default for all reads
        that support the consistency option. It's still possible to override
        this by passing explicitly for a given request. *consistency* can be
        either 'default', 'consistent' or 'stale'.

        *dc* is the datacenter that this agent will communicate with.
        By default the datacenter of the host is used.

        *verify* is whether to verify the SSL certificate for HTTPS requests

        *cert* client side certificates for HTTPS requests
        """

        # TODO: Status

        if os.getenv("CONSUL_HTTP_ADDR") and not (host or port):
            try:
                host, port = os.getenv("CONSUL_HTTP_ADDR").split(":")  # type: ignore
            except ValueError as err:
                raise ConsulException(
                    f"CONSUL_HTTP_ADDR ({os.getenv('CONSUL_HTTP_ADDR')}) invalid, does not match <host>:<port>"
                ) from err
        if not host:
            host = "127.0.0.1"
        if not port:
            port = 8500

        if scheme is None:
            use_ssl = os.getenv("CONSUL_HTTP_SSL")
            scheme = ("https" if use_ssl.lower() == "true" else "http") if use_ssl else "http"

        if verify is None:
            ssl_verify = os.getenv("CONSUL_HTTP_SSL_VERIFY")
            verify = ssl_verify.lower() == "true" if ssl_verify else True

        self.http = self.http_connect(host, port, scheme, verify, cert)
        self.token = os.getenv("CONSUL_HTTP_TOKEN", token)
        self.scheme = scheme
        self.dc = dc
        assert consistency in (
            "default",
            "consistent",
            "stale",
        ), "consistency must be either default, consistent or state"
        self.consistency = consistency

        self.event = Event(self)
        self.kv = KV(self)
        self.txn = Txn(self)
        self.agent = Agent(self)
        self.catalog = Catalog(self)
        self.health = Health(self)
        self.session = Session(self)
        self.acl = ACL(self)
        self.status = Status(self)
        self.query = Query(self)
        self.coordinate = Coordinate(self)
        self.operator = Operator(self)
        self.connect = Connect(self)

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(
        self, exc_type: Optional[type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self.http.close()

    async def __aexit__(
        self, exc_type: Optional[type[BaseException]], exc: Optional[BaseException], tb: Optional[TracebackType]
    ) -> None:
        await self.http.close()

    @abc.abstractmethod
    def http_connect(self, host: str, port: int, scheme, verify: bool = True, cert=None):
        pass

    def prepare_headers(self, token: Optional[str] = None) -> dict[str, str]:
        headers = {}
        if token or self.token:
            headers["X-Consul-Token"] = token or self.token
        return headers  # type: ignore



import time

import pytest
from packaging import version

import consul.check
from tests.utils import should_skip

Check = consul.check.Check


class TestAgent:
    def test_agent_checks(self, consul_port) -> None:
        consul_port, _consul_version = consul_port
        c = consul.Consul(port=consul_port)

        def verify_and_dereg_check(check_id) -> None:
            assert set(c.agent.checks().keys()) == {check_id}
            assert c.agent.check.deregister(check_id) is True
            assert set(c.agent.checks().keys()) == set()

        def verify_check_status(check_id, status, notes=None) -> None:
            checks = c.agent.checks()
            assert checks[check_id]["Status"] == status
            if notes:
                assert checks[check_id]["Output"] == notes

        # test setting notes on a check
        c.agent.check.register("check", Check.ttl("1s"), notes="foo")
        assert c.agent.checks()["check"]["Notes"] == "foo"
        c.agent.check.deregister("check")

        assert set(c.agent.checks().keys()) == set()
        assert c.agent.check.register("script_check", Check.script("/bin/true", 10)) is True
        verify_and_dereg_check("script_check")

        assert c.agent.check.register("check name", Check.script("/bin/true", 10, "10m"), check_id="check_id") is True
        verify_and_dereg_check("check_id")

        # 1s is the minimal interval for HTTP checks
        http_addr = "http://localhost:8500"
        assert c.agent.check.register("http_check", Check.http(http_addr, "1s")) is True
        time.sleep(1.5)
        verify_check_status("http_check", "passing")
        verify_and_dereg_check("http_check")

        assert c.agent.check.register("http_timeout_check", Check.http(http_addr, "100ms", timeout="2s")) is True
        verify_and_dereg_check("http_timeout_check")

        assert c.agent.check.register("ttl_check", Check.ttl("100ms")) is True

        assert c.agent.check.ttl_warn("ttl_check") is True
        verify_check_status("ttl_check", "warning")
        assert c.agent.check.ttl_warn("ttl_check", notes="its not quite right") is True
        verify_check_status("ttl_check", "warning", "its not quite right")

        assert c.agent.check.ttl_fail("ttl_check") is True
        verify_check_status("ttl_check", "critical")
        assert c.agent.check.ttl_fail("ttl_check", notes="something went boink!") is True
        verify_check_status("ttl_check", "critical", notes="something went boink!")

        assert c.agent.check.ttl_pass("ttl_check") is True
        verify_check_status("ttl_check", "passing")
        assert c.agent.check.ttl_pass("ttl_check", notes="all hunky dory!") is True
        verify_check_status("ttl_check", "passing", notes="all hunky dory!")
        # wait for ttl to expire
        time.sleep(120 / 1000.0)
        verify_check_status("ttl_check", "critical")
        verify_and_dereg_check("ttl_check")

    def test_service_multi_check(self, consul_port) -> None:
        consul_port, _consul_version = consul_port
        c = consul.Consul(port=consul_port)
        http_addr = "http://127.0.0.1:8500"
        c.agent.service.register(
            "foo1",
            check=Check.http(http_addr, "1s"),
            extra_checks=[
                Check.http(http_addr, "2s"),
                Check.http(http_addr, "3s"),
            ],
        )

        time.sleep(200 / 1000.0)

        _index, nodes = c.health.service("foo1")
        assert {check["ServiceID"] for node in nodes for check in node["Checks"]} == {"foo1", ""}

        assert {check["CheckID"] for node in nodes for check in node["Checks"]} == {
            "service:foo1:1",
            "service:foo1:2",
            "service:foo1:3",
            "serfHealth",
        }
        time.sleep(3.5)

        _index, checks = c.health.checks(service="foo1")
        assert [check["CheckID"] for check in checks] == ["service:foo1:1", "service:foo1:2", "service:foo1:3"]
        assert [check["Status"] for check in checks] == ["passing", "passing", "passing"]

    def test_service_dereg_issue_156(self, consul_port) -> None:
        consul_port, _consul_version = consul_port
        # https://github.com/cablehead/python-consul/issues/156
        service_name = "app#127.0.0.1#3000"
        c = consul.Consul(port=consul_port)
        c.agent.service.register(service_name)

        time.sleep(80 / 1000.0)

        _index, nodes = c.health.service(service_name)
        assert [node["Service"]["ID"] for node in nodes] == [service_name]

        # Clean up tasks
        assert c.agent.service.deregister(service_name) is True

        time.sleep(40 / 1000.0)

        _index, nodes = c.health.service(service_name)
        assert [node["Service"]["ID"] for node in nodes] == []

    def test_agent_checks_service_id(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        c.agent.service.register("foo1")

        time.sleep(40 / 1000.0)

        _index, nodes = c.health.service("foo1")
        assert [node["Service"]["ID"] for node in nodes] == ["foo1"]

        c.agent.check.register("foo", Check.ttl("100ms"), service_id="foo1")
        c.agent.check.register("foo2", Check.ttl("100ms"), service_id="foo1")

        time.sleep(40 / 1000.0)

        _index, nodes = c.health.service("foo1")
        assert {check["ServiceID"] for node in nodes for check in node["Checks"]} == {"foo1", ""}
        assert {check["CheckID"] for node in nodes for check in node["Checks"]} == {"foo", "foo2", "serfHealth"}

        # Clean up tasks
        assert c.agent.check.deregister("foo") is True
        assert c.agent.check.deregister("foo2") is True

        time.sleep(40 / 1000.0)

        assert c.agent.service.deregister("foo1") is True

        time.sleep(40 / 1000.0)

    def test_agent_register_check_no_service_id(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        _index, nodes = c.health.service("foo1")
        assert nodes == []

        if should_skip(_consul_version, ">", "1.11.0"):
            with pytest.raises(consul.std.base.ConsulException):
                c.agent.check.register("foo", Check.ttl("100ms"), service_id="foo1")
        else:
            assert not c.agent.check.register("foo", Check.ttl("100ms"), service_id="foo1")

        time.sleep(40 / 1000.0)

        assert c.agent.checks() == {}

        # Cleanup tasks
        c.agent.check.deregister("foo")

        time.sleep(40 / 1000.0)

    def test_agent_register_enable_tag_override(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        _index, nodes = c.health.service("foo1")
        assert nodes == []

        c.agent.service.register("foo", enable_tag_override=True)

        assert c.agent.services()["foo"]["EnableTagOverride"]
        # Cleanup tasks
        c.agent.check.deregister("foo")

    def test_agent_service_maintenance(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        c.agent.service.register("foo", check=Check.ttl("100ms"))

        time.sleep(40 / 1000.0)

        c.agent.service.maintenance("foo", "true", "test")

        time.sleep(40 / 1000.0)

        checks_pre = c.agent.checks()
        assert "_service_maintenance:foo" in checks_pre
        assert checks_pre["_service_maintenance:foo"]["Notes"] == "test"

        c.agent.service.maintenance("foo", "false")

        time.sleep(40 / 1000.0)

        checks_post = c.agent.checks()
        assert "_service_maintenance:foo" not in checks_post

        # Cleanup
        c.agent.service.deregister("foo")

        time.sleep(40 / 1000.0)

    def test_agent_node_maintenance(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        c.agent.maintenance("true", "test")

        time.sleep(40 / 1000.0)

        checks_pre = c.agent.checks()
        assert "_node_maintenance" in checks_pre
        assert checks_pre["_node_maintenance"]["Notes"] == "test"

        c.agent.maintenance("false")

        time.sleep(40 / 1000.0)

        checks_post = c.agent.checks()
        assert "_node_maintenance" not in checks_post

    def test_agent_members(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        members = c.agent.members()
        for x in members:
            assert x["Status"] == 1
            assert x["Name"] is not None
            assert x["Tags"] is not None
        assert c.agent.self()["Member"] in members

        wan_members = c.agent.members(wan=True)
        for x in wan_members:
            assert "dc1" in x["Name"]

    def test_agent_self(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        EXPECTED = {
            "v1": {"Member", "Stats", "Config", "Coord", "DebugConfig", "Meta"},
            "v2": {"Member", "xDS", "Stats", "Config", "Coord", "DebugConfig", "Meta"},
        }
        expected = EXPECTED["v1"]
        if version.parse(_consul_version) >= version.parse("1.13.8"):
            expected = EXPECTED["v2"]
        assert set(c.agent.self().keys()) == expected

    def test_agent_services(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        assert c.agent.service.register("foo") is True
        assert set(c.agent.services().keys()) == {"foo"}
        assert c.agent.service.deregister("foo") is True
        assert set(c.agent.services().keys()) == set()

        # test address param
        assert c.agent.service.register("foo", address="10.10.10.1") is True
        assert [v["Address"] for k, v in c.agent.services().items() if k == "foo"][0] == "10.10.10.1"
        assert c.agent.service.deregister("foo") is True



import struct

import pytest

from consul import ConsulException


class TestConsul:
    def test_kv(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        _index, data = c.kv.get("foo")
        assert data is None
        assert c.kv.put("foo", "bar") is True
        _index, data = c.kv.get("foo")
        assert data["Value"] == b"bar"

    def test_kv_wait(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        assert c.kv.put("foo", "bar") is True
        index, _data = c.kv.get("foo")
        check, _data = c.kv.get("foo", index=index, wait="20ms")
        assert index == check

    def test_kv_encoding(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        # test binary
        c.kv.put("foo", struct.pack("i", 1000))
        _index, data = c.kv.get("foo")
        assert struct.unpack("i", data["Value"]) == (1000,)

        # test unicode
        c.kv.put("foo", "bar")
        _index, data = c.kv.get("foo")
        assert data["Value"] == b"bar"

        # test empty-string comes back as `None`
        c.kv.put("foo", "")
        _index, data = c.kv.get("foo")
        assert data["Value"] is None

        # test None
        c.kv.put("foo", None)
        _index, data = c.kv.get("foo")
        assert data["Value"] is None

        # check unencoded values raises assert
        pytest.raises(AssertionError, c.kv.put, "foo", {1: 2})

    def test_kv_put_cas(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        assert c.kv.put("foo", "bar", cas=50) is False
        assert c.kv.put("foo", "bar", cas=0) is True
        _index, data = c.kv.get("foo")

        assert c.kv.put("foo", "bar2", cas=data["ModifyIndex"] - 1) is False
        assert c.kv.put("foo", "bar2", cas=data["ModifyIndex"]) is True
        _index, data = c.kv.get("foo")
        assert data["Value"] == b"bar2"

    def test_kv_put_flags(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        c.kv.put("foo", "bar")
        _index, data = c.kv.get("foo")
        assert data["Flags"] == 0

        assert c.kv.put("foo", "bar", flags=50) is True
        _index, data = c.kv.get("foo")
        assert data["Flags"] == 50

    def test_kv_recurse(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        _index, data = c.kv.get("foo/", recurse=True)
        assert data is None

        c.kv.put("foo/", None)
        _index, data = c.kv.get("foo/", recurse=True)
        assert len(data) == 1

        c.kv.put("foo/bar1", "1")
        c.kv.put("foo/bar2", "2")
        c.kv.put("foo/bar3", "3")
        _index, data = c.kv.get("foo/", recurse=True)
        assert [x["Key"] for x in data] == ["foo/", "foo/bar1", "foo/bar2", "foo/bar3"]
        assert [x["Value"] for x in data] == [None, b"1", b"2", b"3"]

    def test_kv_delete(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        c.kv.put("foo1", "1")
        c.kv.put("foo2", "2")
        c.kv.put("foo3", "3")
        _index, data = c.kv.get("foo", recurse=True)
        assert [x["Key"] for x in data] == ["foo1", "foo2", "foo3"]

        assert c.kv.delete("foo2") is True
        _index, data = c.kv.get("foo", recurse=True)
        assert [x["Key"] for x in data] == ["foo1", "foo3"]
        assert c.kv.delete("foo", recurse=True) is True
        _index, data = c.kv.get("foo", recurse=True)
        assert data is None

    def test_kv_delete_cas(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        c.kv.put("foo", "bar")
        index, data = c.kv.get("foo")

        assert c.kv.delete("foo", cas=data["ModifyIndex"] - 1) is False
        assert c.kv.get("foo") == (index, data)

        assert c.kv.delete("foo", cas=data["ModifyIndex"]) is True
        index, data = c.kv.get("foo")
        assert data is None

    def test_kv_acquire_release(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        pytest.raises(ConsulException, c.kv.put, "foo", "bar", acquire="foo")

        s1 = c.session.create()
        s2 = c.session.create()

        assert c.kv.put("foo", "1", acquire=s1) is True
        assert c.kv.put("foo", "2", acquire=s2) is False
        assert c.kv.put("foo", "1", acquire=s1) is True
        assert c.kv.put("foo", "1", release="foo") is False
        assert c.kv.put("foo", "2", release=s2) is False
        assert c.kv.put("foo", "2", release=s1) is True

        c.session.destroy(s1)
        c.session.destroy(s2)

    def test_kv_keys_only(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        assert c.kv.put("bar", "4") is True
        assert c.kv.put("base/foo", "1") is True
        assert c.kv.put("base/base/foo", "5") is True

        _index, data = c.kv.get("base/", keys=True, separator="/")
        assert data == ["base/base/", "base/foo"]



class TestCoordinates:
    def test_coordinate(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        c.coordinate.nodes()
        c.coordinate.datacenters()
        assert set(c.coordinate.datacenters()[0].keys()) == {"Datacenter", "Coordinates", "AreaID"}



from __future__ import annotations

import base64
import json
from typing import TYPE_CHECKING, Callable

from consul.exceptions import ACLDisabled, ACLPermissionDenied, BadRequest, ClientError, ConsulException, NotFound

if TYPE_CHECKING:
    from consul.base import Response

#
# Conveniences to create consistent callback handlers for endpoints


class CB:
    @classmethod
    def _status(cls, response: Response, allow_404: bool = True) -> None:
        # status checking
        if 400 <= response.code < 500:
            if response.code == 400:
                raise BadRequest(f"{response.code} {response.body}")
            if response.code == 401:
                raise ACLDisabled(response.body)
            if response.code == 403:
                raise ACLPermissionDenied(response.body)
            if response.code == 404:
                if not allow_404:
                    raise NotFound(response.body)
            else:
                raise ClientError(f"{response.code} {response.body}")
        elif 500 <= response.code < 600:
            raise ConsulException(f"{response.code} {response.body}")

    @classmethod
    def boolean(cls) -> Callable[[Response], bool]:
        # returns True on successful response
        def cb(response):
            CB._status(response)
            return response.code == 200

        return cb

    @classmethod
    def json(
        cls,
        postprocess=None,
        allow_404: bool = True,
        one: bool = False,
        decode: bool | str = False,
        is_id: bool = False,
        index: bool = False,
    ):
        """
        *postprocess* is a function to apply to the final result.

        *allow_404* if set, None will be returned on 404, instead of raising
        NotFound.

        *index* if set, a tuple of index, data will be returned.

        *one* returns only the first item of the list of items. empty lists are
        coerced to None.

        *decode* if specified this key will be base64 decoded.

        *is_id* only the 'ID' field of the json object will be returned.
        """

        def cb(response):
            CB._status(response, allow_404=allow_404)
            if response.code == 404:
                data = None
            else:
                try:
                    data = json.loads(response.body)
                    if decode:
                        for item in data:
                            if item.get(decode) is not None:
                                item[decode] = base64.b64decode(item[decode])
                    if is_id:
                        data = data["ID"]
                    if one and isinstance(data, list):
                        data = data[0] if data else None
                    if postprocess:
                        data = postprocess(data)
                except (json.JSONDecodeError, TypeError, KeyError) as e:
                    raise ConsulException(f"Failed to decode JSON: {response.body} {e}") from e
            if index:
                if "X-Consul-Index" not in response.headers:
                    raise ConsulException(f"Missing index header: {response.headers}")
                return response.headers["X-Consul-Index"], data
            return data

        return cb



class TestCatalog:
    pass
    # def test_catalog(self, consul_obj):
    #     c, _consul_version = consul_obj
    #
    #     # grab the node our server created, so we can ignore it
    #     _, nodes = c.catalog.nodes()
    #     assert len(nodes) == 1
    #     current = nodes[0]
    #
    #     # test catalog.datacenters
    #     assert c.catalog.datacenters() == ["dc1"]
    #
    #     # test catalog.register
    #     pytest.raises(consul.ConsulException, c.catalog.register, "foo", "10.1.10.11", dc="dc2")
    #
    #     assert c.catalog.register("n1", "10.1.10.11", service={"service": "s1"}, check={"name": "c1"}) is True
    #     assert c.catalog.register("n1", "10.1.10.11", service={"service": "s2"}) is True
    #     assert c.catalog.register("n2", "10.1.10.12", service={"service": "s1", "tags": ["master"]}) is True
    #
    #     # test catalog.nodes
    #     pytest.raises(consul.ConsulException, c.catalog.nodes, dc="dc2")
    #     _, nodes = c.catalog.nodes()
    #     nodes.remove(current)
    #     assert [x["Node"] for x in nodes] == ["n1", "n2"]
    #
    #     # test catalog.services
    #     pytest.raises(consul.ConsulException, c.catalog.services, dc="dc2")
    #     _, services = c.catalog.services()
    #     assert services == {"s1": ["master"], "s2": [], "consul": []}
    #
    #     # test catalog.node
    #     pytest.raises(consul.ConsulException, c.catalog.node, "n1", dc="dc2")
    #     _, node = c.catalog.node("n1")
    #     assert set(node["Services"].keys()) == {"s1", "s2"}
    #     _, node = c.catalog.node("n3")
    #     assert node is None
    #
    #     # test catalog.service
    #     pytest.raises(consul.ConsulException, c.catalog.service, "s1", dc="dc2")
    #     _, nodes = c.catalog.service("s1")
    #     assert {x["Node"] for x in nodes} == {"n1", "n2"}
    #     _, nodes = c.catalog.service("s1", tag="master")
    #     assert {x["Node"] for x in nodes} == {"n2"}
    #
    #     # test catalog.deregister
    #     pytest.raises(consul.ConsulException, c.catalog.deregister, "n2", dc="dc2")
    #     assert c.catalog.deregister("n1", check_id="c1") is True
    #     assert c.catalog.deregister("n2", service_id="s1") is True
    #     # check the nodes weren't removed
    #     _, nodes = c.catalog.nodes()
    #     nodes.remove(current)
    #     assert [x["Node"] for x in nodes] == ["n1", "n2"]
    #     # check n2's s1 service was removed though
    #     _, nodes = c.catalog.service("s1")
    #     assert {x["Node"] for x in nodes} == {"n1"}
    #
    #     # cleanup
    #     assert c.catalog.deregister("n1") is True
    #     assert c.catalog.deregister("n2") is True
    #     _, nodes = c.catalog.nodes()
    #     nodes.remove(current)
    #     assert [x["Node"] for x in nodes] == []



from __future__ import annotations

from typing import Any

from consul.callback import CB


class Connect:
    def __init__(self, agent) -> None:
        self.agent = agent
        self.ca = Connect.CA(agent)

    class CA:
        def __init__(self, agent) -> None:
            self.agent = agent

        def roots(self, pem: bool = False, token: str | None = None):
            params: list[tuple[str, Any]] = []
            params.append(("pem", int(pem)))

            headers = self.agent.prepare_headers(token)
            return self.agent.http.get(CB.json(), "/v1/connect/ca/roots", params=params, headers=headers)

        def configuration(self, token: str | None = None):
            headers = self.agent.prepare_headers(token)
            return self.agent.http.get(CB.json(), "/v1/connect/ca/configuration", headers=headers)



import pytest

import consul


class TestSession:
    def test_session(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        # session.create
        pytest.raises(consul.ConsulException, c.session.create, node="n2")
        pytest.raises(consul.ConsulException, c.session.create, dc="dc2")
        session_id = c.session.create("my-session")

        # session.list
        pytest.raises(consul.ConsulException, c.session.list, dc="dc2")
        _, sessions = c.session.list()
        assert [x["Name"] for x in sessions] == ["my-session"]

        # session.info
        pytest.raises(consul.ConsulException, c.session.info, session_id, dc="dc2")
        _index, session = c.session.info("1" * 36)
        assert session is None
        _index, session = c.session.info(session_id)
        assert session["Name"] == "my-session"

        # session.node
        node = session["Node"]
        pytest.raises(consul.ConsulException, c.session.node, node, dc="dc2")
        _, sessions = c.session.node(node)
        assert [x["Name"] for x in sessions] == ["my-session"]

        # session.destroy
        pytest.raises(consul.ConsulException, c.session.destroy, session_id, dc="dc2")
        assert c.session.destroy(session_id) is True
        _, sessions = c.session.list()
        assert sessions == []

    def test_session_delete_ttl_renew(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        s = c.session.create(behavior="delete", ttl=20)

        # attempt to renew an unknown session
        pytest.raises(consul.NotFound, c.session.renew, "1" * 36)

        session = c.session.renew(s)
        assert session["Behavior"] == "delete"
        assert session["TTL"] == "20s"

        # trying out the behavior
        assert c.kv.put("foo", "1", acquire=s) is True
        _index, data = c.kv.get("foo")
        assert data["Value"] == b"1"

        c.session.destroy(s)
        _index, data = c.kv.get("foo")
        assert data is None



import asyncio
import base64
import struct

import pytest
from packaging import version

import consul
import consul.aio
import consul.check

Check = consul.check.Check


@pytest.fixture
async def consul_obj(consul_port):
    consul_port, consul_version = consul_port
    c = consul.aio.Consul(port=consul_port)
    yield c, consul_version
    await c.close()


@pytest.fixture
async def consul_acl_obj(acl_consul):
    consul, token, consul_version = acl_consul
    consul.token = token
    yield consul, consul_version
    await consul.close()


class TestAsyncioConsul:
    async def test_kv(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        _index, data = await c.kv.get("foo")
        assert data is None
        response = await c.kv.put("foo", "bar")
        assert response is True
        _index, data = await c.kv.get("foo")
        assert data["Value"] == b"bar"

    async def test_consul_ctor(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        await c.kv.put("foo", struct.pack("i", 1000))
        _index, data = await c.kv.get("foo")
        assert struct.unpack("i", data["Value"]) == (1000,)

    async def test_kv_binary(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        await c.kv.put("foo", struct.pack("i", 1000))
        _index, data = await c.kv.get("foo")
        assert struct.unpack("i", data["Value"]) == (1000,)

    async def test_kv_missing(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        async def put() -> None:
            await asyncio.sleep(2.0 / 100)
            await c.kv.put("foo", "bar")

        fut = asyncio.ensure_future(put())
        await c.kv.put("index", "bump")
        index, data = await c.kv.get("foo")
        assert data is None
        index, data = await c.kv.get("foo", index=index)
        assert data["Value"] == b"bar"
        await fut
        await c.close()

    async def test_kv_put_flags(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        await c.kv.put("foo", "bar")
        _index, data = await c.kv.get("foo")
        assert data["Flags"] == 0

        response = await c.kv.put("foo", "bar", flags=50)
        assert response is True
        _index, data = await c.kv.get("foo")
        assert data["Flags"] == 50

    async def test_kv_delete(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        await c.kv.put("foo1", "1")
        await c.kv.put("foo2", "2")
        await c.kv.put("foo3", "3")
        _index, data = await c.kv.get("foo", recurse=True)
        assert [x["Key"] for x in data] == ["foo1", "foo2", "foo3"]

        response = await c.kv.delete("foo2")
        assert response is True
        _index, data = await c.kv.get("foo", recurse=True)
        assert [x["Key"] for x in data] == ["foo1", "foo3"]
        response = await c.kv.delete("foo", recurse=True)
        assert response is True
        _index, data = await c.kv.get("foo", recurse=True)
        assert data is None

    async def test_kv_subscribe(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        async def put() -> None:
            await asyncio.sleep(1.0 / 100)
            response = await c.kv.put("foo", "bar")
            assert response is True

        fut = asyncio.ensure_future(put())
        index, data = await c.kv.get("foo")
        assert data is None
        index, data = await c.kv.get("foo", index=index)
        assert data["Value"] == b"bar"
        await fut

    async def test_transaction(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        value = base64.b64encode(b"1").decode("utf8")
        d = {"KV": {"Verb": "set", "Key": "asdf", "Value": value}}
        r = await c.txn.put([d])
        assert r["Errors"] is None

        d = {"KV": {"Verb": "get", "Key": "asdf"}}
        r = await c.txn.put([d])
        assert r["Results"][0]["KV"]["Value"] == value

    async def test_agent_services(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        EXPECTED = {
            "v1": {
                "foo": {
                    "Port": 0,
                    "ID": "foo",
                    "CreateIndex": 0,
                    "ModifyIndex": 0,
                    "EnableTagOverride": False,
                    "Service": "foo",
                    "Tags": [],
                    "Meta": {},
                    "Address": "",
                }
            },
            "v2": {
                "foo": {
                    "Address": "",
                    "Datacenter": "dc1",
                    "EnableTagOverride": False,
                    "ID": "foo",
                    "Meta": {},
                    "Port": 0,
                    "Service": "foo",
                    "Tags": [],
                    "Weights": {"Passing": 1, "Warning": 1},
                }
            },
        }
        expected = EXPECTED["v1"]
        if version.parse(_consul_version) >= version.parse("1.13.8"):
            expected = EXPECTED["v2"]

        services = await c.agent.services()
        assert services == {}
        response = await c.agent.service.register("foo")
        assert response is True
        services = await c.agent.services()

        assert services == expected
        response = await c.agent.service.deregister("foo")
        assert response is True
        services = await c.agent.services()
        assert services == {}

    # async def test_catalog(self, acl_consul):
    #     port, token, _consul_version = acl_consul
    #     if should_skip(_consul_version, "<", "1.11.0"):
    #         pytest.skip("Endpoint /v1/acl/create for the legacy ACL system was removed in Consul 1.11.")
    #
    #     c = consul.aio.Consul(port=port, token=token)
    #
    #     async def register():
    #         await asyncio.sleep(1.0 / 100)
    #         response = await c.catalog.register("n1", "10.1.10.11")
    #         assert response is True
    #         await asyncio.sleep(500 / 1000.0)
    #         response = await c.catalog.deregister("n1")
    #         assert response is True
    #
    #     fut = asyncio.ensure_future(register())
    #     index, nodes = await c.catalog.nodes()
    #     assert len(nodes) == 1
    #     current = nodes[0]
    #
    #     index, nodes = await c.catalog.nodes(index=index)
    #     nodes.remove(current)
    #     assert [x["Node"] for x in nodes] == ["n1"]
    #
    #     index, nodes = await c.catalog.nodes(index=index)
    #     nodes.remove(current)
    #     assert [x["Node"] for x in nodes] == []
    #     await fut

    async def test_session(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        async def register() -> None:
            await asyncio.sleep(1.0 / 100)
            session_id = await c.session.create()
            await asyncio.sleep(50 / 1000.0)
            response = await c.session.destroy(session_id)
            assert response is True

        fut = asyncio.ensure_future(register())
        index, services = await c.session.list()
        assert services == []
        await asyncio.sleep(20 / 1000.0)

        index, services = await c.session.list(index=index)
        assert len(services)

        index, services = await c.session.list(index=index)
        assert services == []
        await fut

    # async def test_acl_old(self, acl_consul):
    #     port, token, _consul_version = acl_consul
    #     if should_skip(_consul_version, "<", "1.11.0"):
    #         pytest.skip("Endpoint /v1/acl/create for the legacy ACL system was removed in Consul 1.11.")
    #
    #     c = consul.aio.Consul(port=port, token=token)
    #
    #     rules = """
    #         key "" {
    #             policy = "read"
    #         }
    #         key "private/" {
    #             policy = "deny"
    #         }
    #     """
    #
    #     token = await c.acl.create(rules=rules)
    #
    #     with pytest.raises(consul.ACLPermissionDenied):
    #         await c.acl.list(token=token)
    #
    #     destroyed = await c.acl.destroy(token)
    #     assert destroyed is True



import pytest

import consul
from tests.utils import find_recursive


class TestConsulAcl:
    def test_acl_token_permission_denied(self, acl_consul) -> None:
        c, _master_token, _consul_version = acl_consul

        # No token
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.list)
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.create)
        pytest.raises(
            consul.ACLPermissionDenied, c.acl.token.update, accessor_id="00000000-0000-0000-0000-000000000002"
        )
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.clone, accessor_id="00000000-0000-0000-0000-000000000002")
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.read, accessor_id="00000000-0000-0000-0000-000000000002")
        pytest.raises(
            consul.ACLPermissionDenied, c.acl.token.delete, accessor_id="00000000-0000-0000-0000-000000000002"
        )

        # Token without the right permission (acl:write or acl:read)
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.list, token="anonymous")
        pytest.raises(consul.ACLPermissionDenied, c.acl.token.create, token="anonymous")
        pytest.raises(
            consul.ACLPermissionDenied,
            c.acl.token.update,
            accessor_id="00000000-0000-0000-0000-000000000002",
            token="anonymous",
        )
        pytest.raises(
            consul.ACLPermissionDenied,
            c.acl.token.clone,
            accessor_id="00000000-0000-0000-0000-000000000002",
            token="anonymous",
        )
        pytest.raises(
            consul.ACLPermissionDenied,
            c.acl.token.read,
            accessor_id="00000000-0000-0000-0000-000000000002",
            token="anonymous",
        )
        pytest.raises(
            consul.ACLPermissionDenied,
            c.acl.token.delete,
            accessor_id="00000000-0000-0000-0000-000000000002",
            token="anonymous",
        )

    def test_acl_token_list(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        # Make sure both master and anonymous tokens are created
        acls = c.acl.token.list(token=master_token)

        master_token_repr = {
            "Description": "Initial Management Token",
            "Policies": [{"ID": "00000000-0000-0000-0000-000000000001", "Name": "global-management"}],
            "SecretID": master_token,
        }
        anonymous_token_repr = {
            "AccessorID": "00000000-0000-0000-0000-000000000002",
            "SecretID": "anonymous",
        }
        assert find_recursive(acls, master_token_repr)
        assert find_recursive(acls, anonymous_token_repr)

    def test_acl_token_read(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        # Unknown token
        pytest.raises(consul.ConsulException, c.acl.token.read, accessor_id="unknown", token=master_token)

        anonymous_token_repr = {
            "AccessorID": "00000000-0000-0000-0000-000000000002",
            "SecretID": "anonymous",
        }
        acl = c.acl.token.read(accessor_id="00000000-0000-0000-0000-000000000002", token=master_token)
        assert find_recursive(acl, anonymous_token_repr)

    def test_acl_token_create(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        c.acl.token.create(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        c.acl.token.create(secret_id="DEADBEEF-0000-0000-0000-000000000000", token=master_token)
        c.acl.token.create(
            secret_id="00000000-A5A5-0000-0000-000000000000",
            accessor_id="00000000-0000-A5A5-0000-000000000000",
            description="some token!",
            token=master_token,
        )

        assert c.acl.token.read(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        assert c.acl.token.read(accessor_id="00000000-0000-A5A5-0000-000000000000", token=master_token)

        expected = [
            {
                "AccessorID": "00000000-DEAD-BEEF-0000-000000000000",
                "Description": "",
            },
            {
                "SecretID": "DEADBEEF-0000-0000-0000-000000000000",
                "Description": "",
            },
            {
                "AccessorID": "00000000-0000-A5A5-0000-000000000000",
                "SecretID": "00000000-A5A5-0000-0000-000000000000",
                "Description": "some token!",
            },
        ]
        acl = c.acl.token.list(token=master_token)
        assert find_recursive(acl, expected)

    def test_acl_token_clone(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        assert len(c.acl.token.list(token=master_token)) == 2

        # Unknown token
        pytest.raises(consul.ConsulException, c.acl.token.clone, accessor_id="unknown", token=master_token)

        c.acl.token.create(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        c.acl.token.clone(accessor_id="00000000-DEAD-BEEF-0000-000000000000", description="cloned", token=master_token)
        assert len(c.acl.token.list(token=master_token)) == 4

        expected = [
            {
                "AccessorID": "00000000-DEAD-BEEF-0000-000000000000",
            },
            {
                "Description": "cloned",
            },
        ]
        acl = c.acl.token.list(token=master_token)
        assert find_recursive(acl, expected)

    def test_acl_token_update(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        # Unknown token
        pytest.raises(consul.ConsulException, c.acl.token.update, accessor_id="unknown", token=master_token)

        assert len(c.acl.token.list(token=master_token)) == 2
        c.acl.token.create(
            accessor_id="00000000-DEAD-BEEF-0000-000000000000", description="original", token=master_token
        )
        assert len(c.acl.token.list(token=master_token)) == 3
        c.acl.token.update(
            accessor_id="00000000-DEAD-BEEF-0000-000000000000", description="updated", token=master_token
        )
        assert len(c.acl.token.list(token=master_token)) == 3

        expected = {
            "AccessorID": "00000000-DEAD-BEEF-0000-000000000000",
            "Description": "updated",
        }
        acl = c.acl.token.read(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        assert find_recursive(acl, expected)

    def test_acl_token_delete(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        assert len(c.acl.token.list(token=master_token)) == 2
        c.acl.token.create(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        assert len(c.acl.token.list(token=master_token)) == 3
        assert c.acl.token.read(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)

        # Delete and ensure it doesn't exist anymore
        c.acl.token.delete(accessor_id="00000000-DEAD-BEEF-0000-000000000000", token=master_token)
        assert len(c.acl.token.list(token=master_token)) == 2
        pytest.raises(
            consul.ConsulException,
            c.acl.token.read,
            accessor_id="00000000-DEAD-BEEF-0000-000000000000",
            token=master_token,
        )

    def test_acl_policy_list(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        # Make sure both master and anonymous tokens are created
        policies = c.acl.policy.list(token=master_token)
        assert find_recursive(policies, {"ID": "00000000-0000-0000-0000-000000000001", "Name": "global-management"})

    def test_acl_policy_read(self, acl_consul) -> None:
        c, master_token, _consul_version = acl_consul

        # Unknown token
        pytest.raises(consul.ConsulException, c.acl.policy.read, uuid="unknown", token=master_token)

        policy = c.acl.policy.read(uuid="00000000-0000-0000-0000-000000000001", token=master_token)
        assert find_recursive(policy, {"ID": "00000000-0000-0000-0000-000000000001", "Name": "global-management"})

    #
    # def test_acl_token_implicit_token_use(self, acl_consul):
    #     # configure client to use the master token by default
    #     port, _token, _consul_version = acl_consul
    #     c = consul.Consul(port=port)
    #     master_token = acl_consul.token
    #
    #     if should_skip(_consul_version, "<", "1.11.0"):
    #         clean_consul(port)
    #         pytest.skip("Endpoint /v1/acl/list for the legacy ACL system was removed in Consul 1.11.")
    #
    #
    #     acls = c.acl.list()
    #     assert {x["ID"] for x in acls} == {"anonymous", master_token}
    #
    #     assert c.acl.info("foo") is None
    #     compare = [c.acl.info(master_token), c.acl.info("anonymous")]
    #     compare.sort(key=operator.itemgetter("ID"))
    #     assert acls == compare
    #
    #     rules = """
    #         key "" {
    #             policy = "read"
    #         }
    #         key "private/" {
    #             policy = "deny"
    #         }
    #     """
    #     token = c.acl.create(rules=rules)
    #     assert c.acl.info(token)["Rules"] == rules
    #
    #     token2 = c.acl.clone(token)
    #     assert c.acl.info(token2)["Rules"] == rules
    #
    #     assert c.acl.update(token2, name="Foo") == token2
    #     assert c.acl.info(token2)["Name"] == "Foo"
    #
    #     assert c.acl.destroy(token2) is True
    #     assert c.acl.info(token2) is None
    #
    #     c.kv.put("foo", "bar")
    #     c.kv.put("private/foo", "bar")
    #
    #     c_limited = consul.Consul(port=acl_consul.port, token=token)
    #     assert c_limited.kv.get("foo")[1]["Value"] == b"bar"
    #     pytest.raises(consul.ACLPermissionDenied, c_limited.kv.put, "foo", "bar2")
    #     pytest.raises(consul.ACLPermissionDenied, c_limited.kv.delete, "foo")
    #
    #     assert c.kv.get("private/foo")[1]["Value"] == b"bar"
    #     pytest.raises(consul.ACLPermissionDenied, c_limited.kv.get, "private/foo")
    #     pytest.raises(consul.ACLPermissionDenied, c_limited.kv.put, "private/foo", "bar2")
    #     pytest.raises(consul.ACLPermissionDenied, c_limited.kv.delete, "private/foo")
    #
    #     # check we can override the client's default token
    #     pytest.raises(consul.ACLPermissionDenied, c.kv.get, "private/foo", token=token)
    #     pytest.raises(consul.ACLPermissionDenied, c.kv.put, "private/foo", "bar2", token=token)
    #     pytest.raises(consul.ACLPermissionDenied, c.kv.delete, "private/foo", token=token)
    #
    #     # clean up
    #     c.acl.destroy(token)
    #     acls = c.acl.list()
    #     assert {x["ID"] for x in acls} == {"anonymous", master_token}



import base64


class TestTxn:
    def test_transaction(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        value = base64.b64encode(b"1").decode("utf8")
        d = {"KV": {"Verb": "set", "Key": "asdf", "Value": value}}
        r = c.txn.put([d])
        assert r["Errors"] is None

        d = {"KV": {"Verb": "get", "Key": "asdf"}}
        r = c.txn.put([d])
        assert r["Results"][0]["KV"]["Value"] == value



class TestEvent:
    def test_event(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        assert c.event.fire("fooname", "foobody")
        _index, events = c.event.list()
        assert [x["Name"] == "fooname" for x in events]
        assert [x["Payload"] == "foobody" for x in events]

    def test_event_targeted(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        assert c.event.fire("fooname", "foobody")
        _index, events = c.event.list(name="othername")
        assert events == []

        _index, events = c.event.list(name="fooname")
        assert [x["Name"] == "fooname" for x in events]
        assert [x["Payload"] == "foobody" for x in events]



from __future__ import annotations

import collections
import json
import os
import socket
import time
import uuid

import docker
import pytest
import requests
from docker import DockerClient
from docker.errors import APIError, NotFound
from requests import RequestException

CONSUL_VERSIONS = ["1.17.3", "1.19.2", "1.20.2"]

ConsulInstance = collections.namedtuple("ConsulInstance", ["container", "port", "version"])

# Create a logs directory if it doesn't exist
LOGS_DIR = os.path.join(os.path.dirname(__file__), "logs")
os.makedirs(LOGS_DIR, exist_ok=True)


def get_free_ports(num: int, host=None) -> list[int]:
    if not host:
        host = "127.0.0.1"
    sockets = []
    ret = []
    for _ in range(num):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, 0))
        ret.append(s.getsockname()[1])
        sockets.append(s)
    for s in sockets:
        s.close()
    return ret


@pytest.fixture(scope="session", autouse=True)
def _unset_consul_token():
    if "CONSUL_HTTP_TOKEN" in os.environ:
        del os.environ["CONSUL_HTTP_TOKEN"]


def start_consul_container(version: str, acl_master_token: str | None = None):
    """
    Starts a Consul container. If acl_master_token is None, ACL will be disabled
    for this server, otherwise it will be enabled and the master token will be
    set to the supplied token.

    Returns: a tuple of the container object and the HTTP port the instance is listening on
    """
    client = docker.from_env()
    allocated_ports = get_free_ports(5)
    ports = {
        "http": allocated_ports[0],
        "server": allocated_ports[1],
        "grpc": allocated_ports[2],
        "serf_lan": allocated_ports[3],
        "serf_wan": allocated_ports[4],
    }

    base_config = {
        "ports": {
            "https": -1,
            "dns": -1,
            "grpc_tls": -1,
        },
        "performance": {"raft_multiplier": 1},
        "enable_script_checks": True,
    }
    docker_config = {
        "ports": {
            8500: ports["http"],
            8300: ports["server"],
            8502: ports["grpc"],
            8301: ports["serf_lan"],
            8302: ports["serf_wan"],
        },
        "environment": {"CONSUL_LOCAL_CONFIG": json.dumps(base_config)},
        "detach": True,
        "name": f"consul_test_{uuid.uuid4().hex[:8]}",  # Add a unique name
    }

    # Extend the base config with required ACL fields if needed
    if acl_master_token:
        acl_config = {
            "primary_datacenter": "dc1",
            "acl": {"enabled": True, "tokens": {"initial_management": acl_master_token}},
        }
        merged_config = {**base_config, **acl_config}
        docker_config["environment"]["CONSUL_LOCAL_CONFIG"] = json.dumps(merged_config)  # type: ignore

    def start_consul_container_with_retry(  # pylint: disable=inconsistent-return-statements
        client: DockerClient,
        command: str,
        version: str,
        docker_config: dict,
        max_retries: int = 3,
        retry_delay: int = 2,
    ):
        """
        Start a Consul container with retries as a few initial attempts sometimes fail.
        """
        for attempt in range(max_retries):
            try:
                container = client.containers.run(f"hashicorp/consul:{version}", command=command, **docker_config)
                return container
            except APIError:
                # Cleanup that stray container as it might cause a naming conflict
                try:
                    container = client.containers.get(docker_config["name"])
                    container.remove(force=True)
                except NotFound:
                    pass
                if attempt == max_retries - 1:
                    raise
                time.sleep(retry_delay)

    container = start_consul_container_with_retry(
        client, command="agent -dev -client=0.0.0.0 -log-level trace", version=version, docker_config=docker_config
    )

    # Wait for Consul to be ready
    base_uri = f"http://127.0.0.1:{ports['http']}/v1/"
    start_time = time.time()
    global_timeout = 10

    while True:
        if time.time() - start_time > global_timeout:
            container.stop()
            container.remove()
            raise TimeoutError("Global timeout reached")
        time.sleep(0.1)
        try:
            response = requests.get(base_uri + "status/leader", timeout=2)
            if response.status_code == 200 and response.json():
                break
        except RequestException:
            continue

    # Additional check to ensure Consul is fully ready
    for _ in range(10):
        try:
            requests.put(base_uri + "agent/service/register", json={"name": "test-service"}, timeout=2)
            response = requests.get(base_uri + "health/service/test-service", timeout=2)
            if response.status_code == 200 and response.json():
                requests.put(base_uri + "agent/service/deregister/test-service", timeout=2)
                return container, ports["http"]
        except RequestException:
            time.sleep(0.5)

    container.stop()
    container.remove()
    raise Exception("Failed to verify Consul startup")  # pylint: disable=broad-exception-raised


def get_consul_version(port: int) -> str:
    base_uri = f"http://127.0.0.1:{port}/v1/"
    response = requests.get(base_uri + "agent/self", timeout=10)
    return response.json()["Config"]["Version"].strip()


def setup_and_teardown_consul(request, version, acl_master_token: str | None = None):
    # Start the container, yield, get container logs, store them in logs/<test_name>.log, stop the container
    container, port = start_consul_container(version=version, acl_master_token=acl_master_token)
    version = get_consul_version(port)
    instance = ConsulInstance(container, port, version)

    yield instance if acl_master_token is None else (instance, acl_master_token)

    logs = container.logs().decode("utf-8")
    log_file = os.path.join(LOGS_DIR, f"{request.node.name}.log")
    with open(log_file, "w", encoding="utf-8") as f:
        f.write(logs)

    container.stop()
    container.remove()


@pytest.fixture(params=CONSUL_VERSIONS)
def consul_instance(request):
    yield from setup_and_teardown_consul(request, version=request.param)


@pytest.fixture(params=CONSUL_VERSIONS)
def acl_consul_instance(request):
    acl_master_token = uuid.uuid4().hex
    yield from setup_and_teardown_consul(request, version=request.param, acl_master_token=acl_master_token)


@pytest.fixture
def consul_port(consul_instance):
    return consul_instance.port, consul_instance.version



from __future__ import annotations

import collections
import contextlib
import json
from typing import Any, Callable, Optional

import pytest

import consul.check

Request = collections.namedtuple("Request", ["method", "path", "params", "headers", "data"])


class HTTPClient:
    def __init__(
        self, host: Optional[str] = None, port: Optional[int] = None, scheme=None, verify: bool = True, cert=None
    ) -> None:
        self.host = host
        self.port = int(port) if port else None
        self.scheme = scheme
        self.verify = verify
        self.cert = cert

    def get(self, callback, path, params=None, headers=None):  # pylint: disable=unused-argument
        return Request("get", path, params, headers, None)

    def put(self, callback, path, params=None, headers=None, data: str = ""):  # pylint: disable=unused-argument
        return Request("put", path, params, headers, data)

    def delete(self, callback, path, params=None, headers=None):  # pylint: disable=unused-argument
        return Request("delete", path, params, headers, None)


class Consul(consul.base.Consul):
    def http_connect(self, host: str, port: int, scheme, verify: bool = True, cert=None):
        return HTTPClient(host=host, port=port, scheme=scheme, verify=verify, cert=None)


def _should_support(c: Consul) -> tuple[Callable[..., Any], ...]:
    return (
        # kv
        lambda **kw: c.kv.get("foo", **kw),
        # catalog
        c.catalog.nodes,
        c.catalog.services,
        lambda **kw: c.catalog.node("foo", **kw),
        lambda **kw: c.catalog.service("foo", **kw),
        # session
        c.session.list,
        lambda **kw: c.session.info("foo", **kw),
        lambda **kw: c.session.node("foo", **kw),
    )


def _should_support_node_meta(c: Consul) -> tuple[Callable[..., Any], ...]:
    return (
        # catalog
        c.catalog.nodes,
        c.catalog.services,
        lambda **kw: c.catalog.service("foo", **kw),
        lambda **kw: c.catalog.register("foo", "bar", **kw),
        # health
        lambda **kw: c.health.service("foo", **kw),
        lambda **kw: c.health.checks("foo", **kw),
        lambda **kw: c.health.state("unknown", **kw),
    )


def _should_support_meta(c: Consul) -> tuple[Callable[..., Any], ...]:
    return (
        # agent
        lambda **kw: c.agent.service.register("foo", **kw),
        lambda **kw: c.agent.service.register("foo", "bar", **kw),
    )


class TestBaseInit:
    """
    Tests that connection arguments are handled
    """

    @pytest.mark.parametrize(
        ("env", "host", "port", "want"),
        [
            (
                None,
                None,
                None,
                {
                    "host": "127.0.0.1",
                    "port": 8500,
                },
            ),
            (
                "127.0.0.1:443",
                None,
                None,
                {
                    "host": "127.0.0.1",
                    "port": 443,
                },
            ),
            (
                None,
                "consul.domain.tld",
                None,
                {
                    "host": "consul.domain.tld",
                    "port": 8500,
                },
            ),
            (
                "consul.domain.tld:443",
                "127.0.0.1",
                None,
                {
                    "host": "127.0.0.1",
                    "port": 8500,
                },
            ),
            (
                "consul.domain.tld:443",
                "127.0.0.1",
                8080,
                {
                    "host": "127.0.0.1",
                    "port": 8080,
                },
            ),
            (
                "bad",
                "127.0.0.1",
                8080,
                {
                    "host": "127.0.0.1",
                    "port": 8080,
                },
            ),
        ],
    )
    def test_base_init_addr(self, monkeypatch, env, host, port, want) -> None:
        if env:
            monkeypatch.setenv("CONSUL_HTTP_ADDR", env)
        else:
            with contextlib.suppress(KeyError):
                monkeypatch.delenv("CONSUL_HTTP_ADDR")

        c = Consul(host=host, port=port)
        assert c.http.host == want["host"]
        assert c.http.port == want["port"]

    @pytest.mark.parametrize(
        ("env", "scheme", "want"),
        [
            ("true", None, "https"),
            ("false", None, "http"),
            (None, "https", "https"),
            (None, "http", "http"),
            ("http", "https", "https"),
        ],
    )
    def test_base_init_scheme(self, monkeypatch, env, scheme, want) -> None:
        if env:
            monkeypatch.setenv("CONSUL_HTTP_SSL", env)
        else:
            with contextlib.suppress(KeyError):
                monkeypatch.delenv("CONSUL_HTTP_SSL")

        c = Consul(scheme=scheme)
        assert c.http.scheme == want

    @pytest.mark.parametrize(
        ("env", "verify", "want"),
        [
            ("true", None, True),
            (None, True, True),
            ("false", True, True),
        ],
    )
    def test_base_init_verify(self, monkeypatch, env, verify, want) -> None:
        if env:
            monkeypatch.setenv("CONSUL_HTTP_SSL_VERIFY", env)
        else:
            with contextlib.suppress(KeyError):
                monkeypatch.delenv("CONSUL_HTTP_SSL_VERIFY")

        c = Consul(verify=verify)
        assert c.http.verify == want


class TestIndex:
    """
    Tests read requests that should support blocking on an index
    """

    def test_index(self) -> None:
        c = Consul()
        for r in _should_support(c):
            assert r().params == []
            assert r(index="5").params == [("index", "5")]


class TestConsistency:
    """
    Tests read requests that should support consistency modes
    """

    def test_explict(self) -> None:
        c = Consul()
        for r in _should_support(c):
            assert r().params == []
            assert r(consistency="default").params == []
            assert r(consistency="consistent").params == [("consistent", "1")]
            assert r(consistency="stale").params == [("stale", "1")]

    def test_implicit(self) -> None:
        c = Consul(consistency="consistent")
        for r in _should_support(c):
            assert r().params == [("consistent", "1")]
            assert r(consistency="default").params == []
            assert r(consistency="consistent").params == [("consistent", "1")]
            assert r(consistency="stale").params == [("stale", "1")]


class TestNodemeta:
    """
    Tests read requests that should support node_meta
    """

    def test_node_meta(self) -> None:
        c = Consul()
        for r in _should_support_node_meta(c):
            assert r().params == []
            assert sorted(r(node_meta={"env": "prod", "net": 1}).params) == sorted([
                ("node-meta", "net:1"),
                ("node-meta", "env:prod"),
            ])


class TestMeta:
    """
    Tests read requests that should support meta
    """

    def test_meta(self) -> None:
        c = Consul()
        for r in _should_support_meta(c):
            d = json.loads(r(meta={"env": "prod", "net": 1}).data)
            assert sorted(d["meta"]) == sorted({"env": "prod", "net": 1})


class TestChecks:
    """
    Check constructor helpers return valid check configurations.
    """

    @pytest.mark.parametrize(
        ("url", "interval", "timeout", "deregister", "header", "want"),
        [
            (
                "http://example.com",
                "10s",
                None,
                None,
                None,
                {
                    "http": "http://example.com",
                    "interval": "10s",
                },
            ),
            (
                "http://example.com",
                "10s",
                "1s",
                None,
                None,
                {
                    "http": "http://example.com",
                    "interval": "10s",
                    "timeout": "1s",
                },
            ),
            (
                "http://example.com",
                "10s",
                None,
                "1m",
                None,
                {
                    "http": "http://example.com",
                    "interval": "10s",
                    "DeregisterCriticalServiceAfter": "1m",
                },
            ),
            (
                "http://example.com",
                "10s",
                "1s",
                "1m",
                None,
                {
                    "http": "http://example.com",
                    "interval": "10s",
                    "timeout": "1s",
                    "DeregisterCriticalServiceAfter": "1m",
                },
            ),
            (
                "http://example.com",
                "10s",
                "1s",
                "1m",
                {"X-Test-Header": ["TestHeaderValue"]},
                {
                    "http": "http://example.com",
                    "interval": "10s",
                    "timeout": "1s",
                    "DeregisterCriticalServiceAfter": "1m",
                    "header": {"X-Test-Header": ["TestHeaderValue"]},
                },
            ),
        ],
    )
    def test_http_check(self, url, interval, timeout, deregister, header, want) -> None:
        ch = consul.check.Check.http(url, interval, timeout=timeout, deregister=deregister, header=header)
        assert ch == want

    @pytest.mark.parametrize(
        ("host", "port", "interval", "timeout", "deregister", "want"),
        [
            (
                "localhost",
                1234,
                "10s",
                None,
                None,
                {
                    "tcp": "localhost:1234",
                    "interval": "10s",
                },
            ),
            (
                "localhost",
                1234,
                "10s",
                "1s",
                None,
                {
                    "tcp": "localhost:1234",
                    "interval": "10s",
                    "timeout": "1s",
                },
            ),
            (
                "localhost",
                1234,
                "10s",
                None,
                "1m",
                {
                    "tcp": "localhost:1234",
                    "interval": "10s",
                    "DeregisterCriticalServiceAfter": "1m",
                },
            ),
            (
                "localhost",
                1234,
                "10s",
                "1s",
                "1m",
                {
                    "tcp": "localhost:1234",
                    "interval": "10s",
                    "timeout": "1s",
                    "DeregisterCriticalServiceAfter": "1m",
                },
            ),
        ],
    )
    def test_tcp_check(self, host: str, port: int, interval, timeout, deregister, want) -> None:
        ch = consul.check.Check.tcp(host, port, interval, timeout=timeout, deregister=deregister)
        assert ch == want

    @pytest.mark.parametrize(
        ("container_id", "shell", "script", "interval", "deregister", "want"),
        [
            (
                "wandering_bose",
                "/bin/sh",
                "/bin/true",
                "10s",
                None,
                {
                    "docker_container_id": "wandering_bose",
                    "shell": "/bin/sh",
                    "script": "/bin/true",
                    "interval": "10s",
                },
            ),
            (
                "wandering_bose",
                "/bin/sh",
                "/bin/true",
                "10s",
                "1m",
                {
                    "docker_container_id": "wandering_bose",
                    "shell": "/bin/sh",
                    "script": "/bin/true",
                    "interval": "10s",
                    "DeregisterCriticalServiceAfter": "1m",
                },
            ),
        ],
    )
    def test_docker_check(self, container_id, shell, script, interval, deregister, want) -> None:
        ch = consul.check.Check.docker(container_id, shell, script, interval, deregister=deregister)
        assert ch == want

    def test_ttl_check(self) -> None:
        ch = consul.check.Check.ttl("1m")
        assert ch == {"ttl": "1m"}



from __future__ import annotations

import json
from typing import Optional

from consul.callback import CB


class Policy:
    def __init__(self, agent) -> None:
        self.agent = agent

    def list(self, token: str | None = None):
        """
        Lists all the active ACL policies. This is a privileged endpoint, and
        requires a management token. *token* will override this client's
        default token.
        Requires a token with acl:read capability. ACLPermissionDenied raised otherwise
        """

        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), "/v1/acl/policies", headers=headers)

    def read(self, uuid, token: str | None = None):
        """
        Returns the policy information for *id*. Requires a token with acl:read capability.
        :param uuid: Specifies the UUID of the policy you look up.
        :param token: token with acl:read capability
        :return: selected Polic information
        """
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), f"/v1/acl/policy/{uuid}", headers=headers)

    def create(self, name: str, token: str | None = None, description: Optional[str] = None, rules=None):
        """
        Create a policy
        This is a privileged endpoint, and requires a token with acl:write.
        :param name: Specifies a name for the ACL policy.
        :param token: token with acl:write capability
        :param description: Free form human-readable description of the policy.
        :param rules: Specifies rules for the ACL policy.
        :return: The cloned token information
        """
        json_data = {"name": name}
        if rules:
            json_data["rules"] = json.dumps(rules)
        if description:
            json_data["Description"] = description
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.json(),
            "/v1/acl/policy",
            headers=headers,
            data=json.dumps(json_data),
        )



from __future__ import annotations

from consul.callback import CB


class Health:
    # TODO: All of the health endpoints support all consistency modes
    def __init__(self, agent) -> None:
        self.agent = agent

    def _service(
        self,
        internal_uri,
        index=None,
        wait=None,
        passing=None,
        tag=None,
        dc=None,
        near=None,
        token: str | None = None,
        node_meta=None,
    ):
        params = []
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        if passing:
            params.append(("passing", "1"))
        if tag is not None:
            if not isinstance(tag, list):
                tag = [tag]
            for tag_item in tag:
                params.append(("tag", tag_item))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if near:
            params.append(("near", near))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), internal_uri, params=params, headers=headers)

    def service(self, service: str, **kwargs):
        """
        Returns a tuple of (*index*, *nodes*)

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *nodes* are the nodes providing the given service.

        Calling with *passing* set to True will filter results to only
        those nodes whose checks are currently passing.

        Calling with *tag* will filter the results by tag, multiple tags
        using list possible.

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.
        """
        internal_uri = f"/v1/health/service/{service}"
        return self._service(internal_uri=internal_uri, **kwargs)

    def connect(self, service, **kwargs):
        """
        Returns a tuple of (*index*, *nodes*) of the nodes providing
        connect-capable *service* in the *dc* datacenter. *dc* defaults
        to the current datacenter of this agent.

        Request arguments and response format are the same as health.service
        """
        internal_uri = f"/v1/health/connect/{service}"
        return self._service(internal_uri=internal_uri, **kwargs)

    def checks(self, service, index=None, wait=None, dc=None, near=None, token: str | None = None, node_meta=None):
        """
        Returns a tuple of (*index*, *checks*) with *checks* being the
        checks associated with the service.

        *service* is the name of the service being checked.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.
        """
        params = []
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if near:
            params.append(("near", near))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), f"/v1/health/checks/{service}", params=params, headers=headers)

    def state(self, name: str, index=None, wait=None, dc=None, near=None, token: str | None = None, node_meta=None):
        """
        Returns a tuple of (*index*, *nodes*)

        *name* is a supported state. From the Consul docs:

            The supported states are any, unknown, passing, warning, or
            critical. The any state is a wildcard that can be used to
            return all checks.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.

        *nodes* are the nodes providing the given service.
        """
        assert name in ["any", "unknown", "passing", "warning", "critical"]
        params = []
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if near:
            params.append(("near", near))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), f"/v1/health/state/{name}", params=params, headers=headers)

    def node(self, node, index=None, wait=None, dc=None, token: str | None = None):
        """
        Returns a tuple of (*index*, *checks*)

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *token* is an optional `ACL token`_ to apply to this request.

        *nodes* are the nodes providing the given service.
        """
        params = []
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))

        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), f"/v1/health/node/{node}", params=params, headers=headers)



class ConsulException(Exception):
    pass


class ACLDisabled(ConsulException):
    pass


class ACLPermissionDenied(ConsulException):
    pass


class NotFound(ConsulException):
    pass


class Timeout(ConsulException):
    pass


class BadRequest(ConsulException):
    pass


class ClientError(ConsulException):
    """Encapsulates 4xx Http error code"""



from __future__ import annotations

import json
from typing import Optional

from consul.callback import CB


class Query:
    def __init__(self, agent) -> None:
        self.agent = agent

    def list(self, dc=None, token: str | None = None):
        """
        Lists all the active  queries. This is a privileged endpoint,
        therefore you will only be able to get the prepared queries
        which the token supplied has read privileges to.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.

        *token* is an optional `ACL token`_ to apply to this request.
        """
        params = []
        if dc:
            params.append(("dc", dc))

        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), "/v1/query", params=params, headers=headers)

    def _query_data(
        self,
        service=None,
        name: Optional[str] = None,
        session=None,
        token: str | None = None,
        nearestn=None,
        datacenters=None,
        onlypassing=None,
        tags=None,
        ttl: Optional[int] = None,
        regexp=None,
    ):
        service_body = {
            k: v
            for k, v in {
                "service": service,
                "onlypassing": onlypassing,
                "tags": tags,
                "failover": {
                    k: v for k, v in {"nearestn": nearestn, "datacenters": datacenters}.items() if v is not None
                },
            }.items()
            if v is not None
        }

        data = {
            k: v
            for k, v in {
                "name": name,
                "session": session,
                "token": token or self.agent.token,
                "dns": {"ttl": ttl} if ttl is not None else None,
                "template": {k: v for k, v in {"type": "name_prefix_match", "regexp": regexp}.items() if v is not None},
                "service": service_body,
            }.items()
            if v is not None
        }
        return json.dumps(data)

    def create(
        self,
        service,
        name: Optional[str] = None,
        dc=None,
        session=None,
        token: str | None = None,
        nearestn=None,
        datacenters=None,
        onlypassing=None,
        tags=None,
        ttl: Optional[int] = None,
        regexp=None,
    ):
        """
        Creates a new query. This is a privileged endpoint, and
        requires a management token for a certain query name.*token* will
        override this client's default token.

        *service* is mandatory for new query. represent service name to
        query.

        *name* is an optional name for this query.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.

        *token* is an optional `ACL token`_ to apply to this request.

        *nearestn* if set to a value greater than zero, then the query will
        be forwarded to up to NearestN other datacenters based on their
        estimated network round trip time using Network Coordinates from
        the WAN gossip pool.

        *datacenters* is a fixed list of remote datacenters to forward the
        query to if there are no healthy nodes in the local datacenter.

        *onlypassing* controls the behavior of the query's health check
        filtering.

        *tags* is a list of service tags to filter the query results.

        *ttl*  is a duration string that can use "s" as a suffix for
        seconds. It controls how the TTL is set when query results are
        served over DNS.

        *regexp* is optional for template this option is only supported
        in Consul 0.6.4 or later. The only option for type is
        name_prefix_match so if you want a query template with no regexp
        enter an empty string.

        For more information about query
        https://www.consul.io/docs/agent/http/query.html
        """
        path = "/v1/query"
        params = None if dc is None else [("dc", dc)]
        data = self._query_data(service, name, session, token, nearestn, datacenters, onlypassing, tags, ttl, regexp)
        return self.agent.http.post(CB.json(), path, params=params, data=data)

    def update(
        self,
        query_id,
        service=None,
        name: Optional[str] = None,
        dc=None,
        session=None,
        token: str | None = None,
        nearestn=None,
        datacenters=None,
        onlypassing=None,
        tags=None,
        ttl: Optional[int] = None,
        regexp=None,
    ):
        """
        This endpoint will update a certain query

        *query_id* is the query id for update

        all the other setting remains the same as the query create method
        """
        path = f"/v1/query/{query_id}"
        params = None if dc is None else [("dc", dc)]
        data = self._query_data(service, name, session, token, nearestn, datacenters, onlypassing, tags, ttl, regexp)
        return self.agent.http.put(CB.boolean(), path, params=params, data=data)

    def get(self, query_id, token: str | None = None, dc=None):
        """
        This endpoint will return information about a certain query

        *query_id* the query id to retrieve information about

        *token* is an optional `ACL token`_ to apply to this request.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.
        """
        params = []
        if dc:
            params.append(("dc", dc))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), f"/v1/query/{query_id}", params=params, headers=headers)

    def delete(self, query_id, token: str | None = None, dc=None):
        """
        This endpoint will delete certain query

        *query_id* the query id delete

        *token* is an optional `ACL token`_ to apply to this request.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.
        """
        params = []
        if dc:
            params.append(("dc", dc))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.delete(CB.boolean(), f"/v1/query/{query_id}", params=params, headers=headers)

    def execute(self, query, token: str | None = None, dc=None, near=None, limit: Optional[int] = None):
        """
        This endpoint will execute certain query

        *query* name or query id to execute

        *token* is an optional `ACL token`_ to apply to this request.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *limit* is used to limit the size of the list to the given number
        of nodes. This is applied after any sorting or shuffling.
        """
        params = []
        if dc:
            params.append(("dc", dc))
        if near:
            params.append(("near", near))
        if limit:
            params.append(("limit", limit))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), f"/v1/query/{query}/execute", params=params, headers=headers)

    def explain(self, query, token: str | None = None, dc=None):
        """
        This endpoint shows a fully-rendered query for a given name

        *query* name to explain. This cannot be query id.

        *token* is an optional `ACL token`_ to apply to this request.

        *dc* is the datacenter that this agent will communicate with. By
        default the datacenter of the host is used.
        """
        params = []
        if dc:
            params.append(("dc", dc))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(), f"/v1/query/{query}/explain", params=params, headers=headers)



from __future__ import annotations

import json

from consul.callback import CB


class Catalog:
    def __init__(self, agent) -> None:
        self.agent = agent

    def register(self, node, address, service=None, check=None, dc=None, token: str | None = None, node_meta=None):
        """
        A low level mechanism for directly registering or updating entries
        in the catalog. It is usually recommended to use
        agent.service.register and agent.check.register, as they are
        simpler and perform anti-entropy.

        *node* is the name of the node to register.

        *address* is the ip of the node.

        *service* is an optional service to register. if supplied this is a
        dict::

            {
                "Service": "redis",
                "ID": "redis1",
                "Tags": [
                    "master",
                    "v1"
                ],
                "Port": 8000
            }

        where

            *Service* is required and is the name of the service

            *ID* is optional, and will be set to *Service* if not provided.
            Note *ID* must be unique for the given *node*.

            *Tags* and *Port* are optional.

        *check* is an optional check to register. if supplied this is a
        dict::

            {
                "Node": "foobar",
                "CheckID": "service:redis1",
                "Name": "Redis health check",
                "Notes": "Script based health check",
                "Status": "passing",
                "ServiceID": "redis1"
            }

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.

        This manipulates the health check entry, but does not setup a
        script or TTL to actually update the status. The full documentation
        is `here <https://consul.io/docs/agent/http.html#catalog>`_.

        Returns *True* on success.
        """
        data = {"node": node, "address": address}
        params = []
        dc = dc or self.agent.dc
        if dc:
            data["datacenter"] = dc
        if service:
            data["service"] = service
        if check:
            data["check"] = check
        token = token or self.agent.token
        if token:
            data["WriteRequest"] = {"Token": token}
            params.append(("token", token))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))

        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(
            CB.boolean(), "/v1/catalog/register", data=json.dumps(data), params=params, headers=headers
        )

    def deregister(self, node, service_id=None, check_id=None, dc=None, token: str | None = None):
        """
        A low level mechanism for directly removing entries in the catalog.
        It is usually recommended to use the agent APIs, as they are
        simpler and perform anti-entropy.

        *node* and *dc* specify which node on which datacenter to remove.
        If *service_id* and *check_id* are not provided, all associated
        services and checks are deleted. Otherwise only one of *service_id*
        and *check_id* should be provided and only that service or check
        will be removed.

        *token* is an optional `ACL token`_ to apply to this request.

        Returns *True* on success.
        """
        assert not (service_id and check_id)
        data = {"node": node}
        dc = dc or self.agent.dc
        if dc:
            data["datacenter"] = dc
        if service_id:
            data["serviceid"] = service_id
        if check_id:
            data["checkid"] = check_id
        token = token or self.agent.token
        if token:
            data["WriteRequest"] = {"Token": token}
        headers = self.agent.prepare_headers(token)
        return self.agent.http.put(CB.boolean(), "/v1/catalog/deregister", headers=headers, data=json.dumps(data))

    def datacenters(self):
        """
        Returns all the datacenters that are known by the Consul server.
        """
        return self.agent.http.get(CB.json(), "/v1/catalog/datacenters")

    def nodes(
        self, index=None, wait=None, consistency=None, dc=None, near=None, token: str | None = None, node_meta=None
    ):
        """
        Returns a tuple of (*index*, *nodes*) of all nodes known
        about in the *dc* datacenter. *dc* defaults to the current
        datacenter of this agent.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *consistency* can be either 'default', 'consistent' or 'stale'. if
        not specified *consistency* will the consistency level this client
        was configured with.

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.

        The response looks like this::

            (index, [
                {
                    "Node": "baz",
                    "Address": "10.1.10.11"
                },
                {
                    "Node": "foobar",
                    "Address": "10.1.10.12"
                }
            ])
        """
        params = []
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        if near:
            params.append(("near", near))

        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), "/v1/catalog/nodes", params=params, headers=headers)

    def services(self, index=None, wait=None, consistency=None, dc=None, token: str | None = None, node_meta=None):
        """
        Returns a tuple of (*index*, *services*) of all services known
        about in the *dc* datacenter. *dc* defaults to the current
        datacenter of this agent.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *consistency* can be either 'default', 'consistent' or 'stale'. if
        not specified *consistency* will the consistency level this client
        was configured with.

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.

        The response looks like this::

            (index, {
                "consul": [],
                "redis": [],
                "postgresql": [
                    "master",
                    "slave"
                ]
            })

        The main keys are the service names and the list provides all the
        known tags for a given service.
        """
        params = []
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), "/v1/catalog/services", params=params, headers=headers)

    def node(self, node, index=None, wait=None, consistency=None, dc=None, token: str | None = None):
        """
        Returns a tuple of (*index*, *services*) of all services provided
        by *node*.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        *consistency* can be either 'default', 'consistent' or 'stale'. if
        not specified *consistency* will the consistency level this client
        was configured with.

        *dc* is the datacenter of the node and defaults to this agents
        datacenter.

        *token* is an optional `ACL token`_ to apply to this request.

        The response looks like this::

            (index, {
                "Node": {
                    "Node": "foobar",
                    "Address": "10.1.10.12"
                },
                "Services": {
                    "consul": {
                        "ID": "consul",
                        "Service": "consul",
                        "Tags": null,
                        "Port": 8300
                    },
                    "redis": {
                        "ID": "redis",
                        "Service": "redis",
                        "Tags": [
                            "v1"
                        ],
                        "Port": 8000
                    }
                }
            })
        """
        params = []
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), f"/v1/catalog/node/{node}", params=params, headers=headers)

    def _service(
        self,
        internal_uri,
        index=None,
        wait=None,
        tag=None,
        consistency=None,
        dc=None,
        near=None,
        token: str | None = None,
        node_meta=None,
    ):
        params = []
        dc = dc or self.agent.dc
        if dc:
            params.append(("dc", dc))
        if tag:
            params.append(("tag", tag))
        if index:
            params.append(("index", index))
            if wait:
                params.append(("wait", wait))
        if near:
            params.append(("near", near))
        consistency = consistency or self.agent.consistency
        if consistency in ("consistent", "stale"):
            params.append((consistency, "1"))
        if node_meta:
            for nodemeta_name, nodemeta_value in node_meta.items():
                params.append(("node-meta", f"{nodemeta_name}:{nodemeta_value}"))
        headers = self.agent.prepare_headers(token)
        return self.agent.http.get(CB.json(index=True), internal_uri, params=params, headers=headers)

    def service(self, service: str, **kwargs):
        """
        Returns a tuple of (*index*, *nodes*) of the nodes providing
        *service* in the *dc* datacenter. *dc* defaults to the current
        datacenter of this agent.

        *index* is the current Consul index, suitable for making subsequent
        calls to wait for changes since this query was last run.

        *wait* the maximum duration to wait (e.g. '10s') to retrieve
        a given index. this parameter is only applied if *index* is also
        specified. the wait time by default is 5 minutes.

        If *tag* is provided, the list of nodes returned will be filtered
        by that tag.

        *near* is a node name to sort the resulting list in ascending
        order based on the estimated round trip time from that node

        *consistency* can be either 'default', 'consistent' or 'stale'. if
        not specified *consistency* will the consistency level this client
        was configured with.

        *token* is an optional `ACL token`_ to apply to this request.

        *node_meta* is an optional meta data used for filtering, a
        dictionary formatted as {k1:v1, k2:v2}.

        The response looks like this::

            (index, [
                {
                    "Node": "foobar",
                    "Address": "10.1.10.12",
                    "ServiceID": "redis",
                    "ServiceName": "redis",
                    "ServiceTags": null,
                    "ServicePort": 8000
                }
            ])
        """
        internal_uri = f"/v1/catalog/service/{service}"
        return self._service(internal_uri=internal_uri, **kwargs)

    def connect(self, service: str, **kwargs):
        """
        Returns a tuple of (*index*, *nodes*) of the nodes providing
        connect-capable *service* in the *dc* datacenter. *dc* defaults
        to the current datacenter of this agent.

        Request arguments and response format are the same as catalog.service
        """
        internal_uri = f"/v1/catalog/connect/{service}"
        return self._service(internal_uri=internal_uri, **kwargs)



class testQuery:
    def test_query(self, consul_obj) -> None:
        c, _consul_version = consul_obj

        # check that query list is empty
        queries = c.query.list()
        assert queries == []

        # create a new named query
        query_service = "foo"
        query_name = "fooquery"
        query = c.query.create(query_service, query_name)

        # assert response contains query ID
        assert "ID" in query
        assert query["ID"] is not None
        assert str(query["ID"]) != ""

        # retrieve query using id and name
        queries = c.query.get(query["ID"])
        assert queries != []
        assert len(queries) == 1
        assert queries[0]["Name"] == query_name
        assert queries[0]["ID"] == query["ID"]

        # explain query
        assert c.query.explain(query_name)["Query"]

        # delete query
        assert c.query.delete(query["ID"])



import consul
import consul.check
import consul.std


class TestHTTPClient:
    def test_uri(self) -> None:
        http = consul.std.HTTPClient()
        assert http.uri("/v1/kv") == "http://127.0.0.1:8500/v1/kv"
        assert http.uri("/v1/kv", params=[("index", 1)]) == "http://127.0.0.1:8500/v1/kv?index=1"



from typing import Optional

import aiohttp

from consul import Timeout, base

__all__ = ["Consul"]


class HTTPClient(base.HTTPClient):
    """Asyncio adapter for python consul using aiohttp library"""

    def __init__(self, *args, loop=None, connections_limit=None, connections_timeout=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.loop = loop
        connector_kwargs = {}
        if connections_limit:
            connector_kwargs["limit"] = connections_limit
        connector = aiohttp.TCPConnector(loop=self.loop, verify_ssl=self.verify, **connector_kwargs)
        session_kwargs = {}
        if connections_timeout:
            timeout = aiohttp.ClientTimeout(total=connections_timeout)
            session_kwargs["timeout"] = timeout
        self._session = aiohttp.ClientSession(connector=connector, **session_kwargs)  # type: ignore

    async def _request(
        self, callback, method, uri, headers: Optional[dict[str, str]], data=None, connections_timeout=None
    ):
        session_kwargs = {}
        if connections_timeout:
            timeout = aiohttp.ClientTimeout(total=connections_timeout)
            session_kwargs["timeout"] = timeout
        resp = await self._session.request(method, uri, headers=headers, data=data, **session_kwargs)  # type: ignore
        body = await resp.text(encoding="utf-8")
        if resp.status == 599:
            raise Timeout
        r = base.Response(resp.status, resp.headers, body)
        return callback(r)

    def get(self, callback, path, params=None, headers: Optional[dict[str, str]] = None, connections_timeout=None):
        uri = self.uri(path, params)
        return self._request(callback, "GET", uri, headers=headers, connections_timeout=connections_timeout)

    def put(
        self,
        callback,
        path,
        params=None,
        data: str = "",
        headers: Optional[dict[str, str]] = None,
        connections_timeout=None,
    ):
        uri = self.uri(path, params)
        return self._request(callback, "PUT", uri, headers=headers, data=data, connections_timeout=connections_timeout)

    def delete(self, callback, path, params=None, headers: Optional[dict[str, str]] = None, connections_timeout=None):
        uri = self.uri(path, params)
        return self._request(callback, "DELETE", uri, headers=headers, connections_timeout=connections_timeout)

    def post(
        self,
        callback,
        path,
        params=None,
        data: str = "",
        headers: Optional[dict[str, str]] = None,
        connections_timeout=None,
    ):
        uri = self.uri(path, params)
        return self._request(callback, "POST", uri, headers=headers, data=data, connections_timeout=connections_timeout)

    def close(self):
        return self._session.close()


class Consul(base.Consul):
    def __init__(self, *args, loop=None, connections_limit=None, connections_timeout=None, **kwargs) -> None:
        self.loop = loop
        self.connections_limit = connections_limit
        self.connections_timeout = connections_timeout
        super().__init__(*args, **kwargs)

    def http_connect(self, host: str, port: int, scheme, verify: bool = True, cert=None):
        return HTTPClient(
            host,
            port,
            scheme,
            loop=self.loop,
            connections_limit=self.connections_limit,
            connections_timeout=self.connections_timeout,
            verify=verify,
            cert=cert,
        )

    def close(self):
        """Close all opened http connections"""
        return self.http.close()



from consul.api.acl.policy import Policy
from consul.api.acl.token import Token


class ACL:
    def __init__(self, agent) -> None:
        self.agent = agent

        self.token = self.tokens = Token(agent)
        self.policy = self.policies = Policy(agent)



from packaging import version


class TestOperator:
    def test_operator(self, consul_obj) -> None:
        c, _consul_version = consul_obj
        config = c.operator.raft_config()

        expected_index = 1
        if version.parse(_consul_version) >= version.parse("1.13.8"):
            expected_index = 0

        assert config["Index"] == expected_index
        leader = False
        voter = False
        for server in config["Servers"]:
            if server["Leader"]:
                leader = True
            if server["Voter"]:
                voter = True
        assert leader
        assert voter



import glob
import os
import re

from setuptools import find_packages, setup
from setuptools.command.install import install

with open("consul/__init__.py", encoding="utf-8") as f:
    metadata = dict(re.findall('__([a-z]+)__ = "([^"]+)"', f.read()))


def _read_reqs(relpath: str):
    fullpath = os.path.join(os.path.dirname(__file__), relpath)
    with open(fullpath, encoding="utf-8") as f:
        return [s.strip() for s in f.readlines() if (s.strip() and not s.startswith("#"))]


description = "Python client for Consul (http://www.consul.io/)"


py_modules = [os.path.splitext(x)[0] for x in glob.glob("consul/*.py")]


class Install(install):
    def run(self):
        install.run(self)


with open("README.md", encoding="utf-8") as f1, open("CHANGELOG.md", encoding="utf-8") as f2:
    long_description = f"{f1.read()}\n\n{f2.read()}"

setup(
    name="py-consul",
    version=metadata["version"],
    author="Criteo",
    author_email="github@criteo.com",
    url="https://github.com/criteo-forks/py-consul",
    license="MIT",
    description=description,
    description_content_type="text/markdown",
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=py_modules,
    install_requires=_read_reqs("requirements.txt"),
    extras_require={
        "asyncio": ["aiohttp"],
    },
    data_files=[(".", ["requirements.txt", "tests-requirements.txt"])],
    packages=find_packages(exclude=["tests*"]),
    tests_require=_read_reqs("tests-requirements.txt"),
    cmdclass={"install": Install},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)
